from typing import Dict, List, Optional, Tuple

import pytest
from bson.objectid import ObjectId
from inline_snapshot import snapshot
from pydantic import ValidationError, model_validator, root_validator
from pydantic.main import BaseModel

from odmantic.exceptions import DocumentParsingError
from odmantic.field import Field
from odmantic.model import EmbeddedModel, Model
from odmantic.reference import Reference
from tests.integration.utils import redact_objectid
from tests.zoo.person import PersonModel


def test_repr_model():
    class M(Model):
        a: int

    instance = M(a=5)
    assert repr(instance) == f"M(id={repr(instance.id)}, a=5)"


def test_repr_embedded_model():
    class M(EmbeddedModel):
        a: int

    instance = M(a=5)
    assert repr(instance) == "M(a=5)"


def test_fields_modified_no_modification():
    class M(Model):
        f: int

    instance = M(f=0)
    assert instance.__fields_modified__ == set(["f", "id"])


def test_fields_embedded_modified_no_modification():
    class M(EmbeddedModel):
        f: int

    instance = M(f=0)
    assert instance.__fields_modified__ == set(["f"])


def test_fields_modified_with_default():
    class M(Model):
        f: int = 5

    instance = M(f=0)
    assert instance.__fields_modified__ == set(["f", "id"])


@pytest.mark.parametrize("model_cls", [Model, EmbeddedModel])
def test_fields_modified_one_update(model_cls):
    class M(model_cls):  # type: ignore
        f: int

    instance = M(f=0)
    instance.__fields_modified__.clear()
    instance.f = 1
    assert instance.__fields_modified__ == set(["f"])


def test_field_update_with_invalid_data_type():
    class M(Model):
        f: int

    instance = M(f=0)
    with pytest.raises(ValidationError):
        instance.f = "aa"  # type: ignore


def test_field_update_with_invalid_data():
    class M(Model):
        f: int = Field(gt=0)

    instance = M(f=1)
    with pytest.raises(ValidationError):
        instance.f = -1


def test_validate_does_not_copy():
    instance = PersonModel(first_name="Jean", last_name="Pierre")
    assert PersonModel.validate(instance) is instance


def test_validate_from_dict():
    instance = PersonModel.validate({"first_name": "Jean", "last_name": "Pierre"})
    assert isinstance(instance, PersonModel)
    assert instance.first_name == "Jean" and instance.last_name == "Pierre"


def test_fields_modified_on_construction():
    instance = PersonModel(first_name="Jean", last_name="Pierre")
    assert instance.__fields_modified__ == set(["first_name", "last_name", "id"])


def test_fields_modified_on_document_parsing():
    instance = PersonModel.model_validate_doc(
        {"_id": ObjectId(), "first_name": "Jackie", "last_name": "Chan"}
    )
    assert instance.__fields_modified__ == set(["first_name", "last_name", "id"])


def test_document_parsing_error_keyname():
    class M(Model):
        field: str = Field(key_name="custom")

    id = ObjectId()
    with pytest.raises(DocumentParsingError) as exc_info:
        M.model_validate_doc({"_id": id})
    assert redact_objectid(str(exc_info.value), id) == snapshot(
        """\
1 validation error for M
field
  Key 'custom' not found in document [type=odmantic::key_not_found_in_document, input_value={'_id': ObjectId('<ObjectId>')}, input_type=dict]\
"""  # noqa: E501
    )


def test_document_parsing_error_embedded_keyname():
    class F(EmbeddedModel):
        a: int

    class E(EmbeddedModel):
        f: F

    class M(Model):
        e: E

    with pytest.raises(DocumentParsingError) as exc_info:
        M.model_validate_doc({"_id": ObjectId(), "e": {"f": {}}})
    assert str(exc_info.value) == snapshot(
        """\
1 validation error for M
e.f.a
  Key 'a' not found in document [type=odmantic::key_not_found_in_document, input_value={}, input_type=dict]\
"""  # noqa: E501
    )


def test_embedded_document_parsing_error():
    class E(EmbeddedModel):
        f: int

    with pytest.raises(DocumentParsingError) as exc_info:
        E.model_validate_doc({})
    assert str(exc_info.value) == snapshot(
        """\
1 validation error for E
f
  Key 'f' not found in document [type=odmantic::key_not_found_in_document, input_value={}, input_type=dict]\
"""  # noqa: E501
    )


def test_embedded_document_parsing_validation_error():
    class E(EmbeddedModel):
        f: int

    with pytest.raises(DocumentParsingError) as exc_info:
        E.model_validate_doc({"f": "aa"})
    assert str(exc_info.value).splitlines()[:-1] == snapshot(
        [
            "1 validation error for E",
            "f",
            "  Input should be a valid integer, unable to parse string as an integer [type=int_parsing, input_value='aa', input_type=str]",  # noqa: E501
        ]
    )


def test_embedded_model_alternate_key_name_with_default():
    class Em(EmbeddedModel):
        name: str = Field(key_name="username")

    class M(Model):
        f: Em = Em(name="Jack")

    _id = ObjectId()
    doc = {"_id": _id}
    parsed = M.model_validate_doc(doc)
    assert parsed.f.name == "Jack"


def test_embedded_model_alternate_key_name_parsing_exception():
    class Em(EmbeddedModel):
        name: str = Field(key_name="username")

    class M(Model):
        f: Em

    _id = ObjectId()
    doc = {"_id": _id}
    with pytest.raises(DocumentParsingError):
        M.model_validate_doc(doc)


def test_embedded_model_alternate_key_name():
    class Em(EmbeddedModel):
        name: str = Field(key_name="username")

    class M(Model):
        f: Em

    instance = M(f=Em(name="Jack"))
    doc = instance.model_dump_doc()
    assert doc["f"] == {"username": "Jack"}
    parsed = M.model_validate_doc(doc)
    assert parsed == instance


def test_embedded_model_list_alternate_key_name():
    class Em(EmbeddedModel):
        name: str = Field(key_name="username")

    class M(Model):
        f: List[Em]

    instance = M(f=[Em(name="Jack")])
    doc = instance.model_dump_doc()
    assert doc["f"] == [{"username": "Jack"}]
    parsed = M.model_validate_doc(doc)
    assert parsed == instance


def test_embedded_model_tuple_alternate_key_name():
    class Em(EmbeddedModel):
        name: str = Field(key_name="username")

    class M(Model):
        f: Tuple[Em, ...]

    instance = M(f=(Em(name="Jack"),))
    doc = instance.model_dump_doc()
    assert doc["f"] == [{"username": "Jack"}]
    parsed = M.model_validate_doc(doc)
    assert parsed == instance


def test_embedded_model_list_parsing_invalid_type():
    class Em(EmbeddedModel):
        name: str

    class M(Model):
        f: List[Em]

    with pytest.raises(DocumentParsingError) as exc_info:
        M.model_validate_doc({"_id": 1, "f": {1: {"name": "Jack"}}})

    assert str(exc_info.value) == snapshot(
        """\
1 validation error for M
f
  Incorrect generic embedded model value '{1: {'name': 'Jack'}}' [type=odmantic::incorrect_generic_embedded_model_value, input_value={'_id': 1, 'f': {1: {'name': 'Jack'}}}, input_type=dict]\
"""  # noqa: E501
    )


def test_embedded_model_list_parsing_missing_value():
    class Em(EmbeddedModel):
        name: str

    class M(Model):
        f: List[Em]

    with pytest.raises(
        DocumentParsingError,
    ) as exc_info:
        M.model_validate_doc({"_id": 1})
    assert str(exc_info.value) == snapshot(
        """\
1 validation error for M
f
  Key 'f' not found in document [type=odmantic::key_not_found_in_document, input_value={'_id': 1}, input_type=dict]\
"""  # noqa: E501
    )


def test_embedded_model_list_parsing_missing_value_with_default():
    class Em(EmbeddedModel):
        name: str

    class M(Model):
        f: List[Em] = [Em(name="John")]

    parsed = M.model_validate_doc({"_id": ObjectId()})
    assert parsed.f == [Em(name="John")]


def test_embedded_model_dict_parsing_invalid_value():
    class Em(EmbeddedModel):
        name: str

    class M(Model):
        f: Dict[str, Em]

    with pytest.raises(DocumentParsingError) as exc_info:
        M.model_validate_doc({"_id": 1, "f": []})
    assert str(exc_info.value) == snapshot(
        """\
1 validation error for M
f
  Incorrect generic embedded model value '[]' [type=odmantic::incorrect_generic_embedded_model_value, input_value={'_id': 1, 'f': []}, input_type=dict]\
"""  # noqa: E501
    )


def test_embedded_model_dict_parsing_invalid_sub_value():
    class Em(EmbeddedModel):
        e: int

    class M(Model):
        f: Dict[str, Em]

    with pytest.raises(DocumentParsingError) as exc_info:
        M.model_validate_doc({"_id": ObjectId(), "f": {"key": {"not_there": "a"}}})
    assert str(exc_info.value) == snapshot(
        """\
1 validation error for M
f.["key"].e
  Key 'e' not found in document [type=odmantic::key_not_found_in_document, input_value={'not_there': 'a'}, input_type=dict]\
"""  # noqa: E501
    )


def test_embedded_model_list_parsing_invalid_sub_value():
    class Em(EmbeddedModel):
        e: int

    class M(Model):
        f: List[Em]

    with pytest.raises(DocumentParsingError) as exc_info:
        M.model_validate_doc({"_id": ObjectId(), "f": [{"not_there": "a"}]})
    assert str(exc_info.value) == snapshot(
        """\
1 validation error for M
f.[0].e
  Key 'e' not found in document [type=odmantic::key_not_found_in_document, input_value={'not_there': 'a'}, input_type=dict]\
"""  # noqa: E501
    )


def test_fields_modified_on_object_parsing():
    instance = PersonModel.model_validate(
        {"_id": ObjectId(), "first_name": "Jackie", "last_name": "Chan"}
    )
    assert instance.__fields_modified__ == set(["first_name", "last_name", "id"])


def test_change_primary_key_value():
    class M(Model): ...

    instance = M()
    with pytest.raises(NotImplementedError, match="assigning a new primary key"):
        instance.id = 12


def test_model_copy_without_update():
    instance = PersonModel(first_name="Jean", last_name="Valjean")
    copied = instance.model_copy()
    assert instance == copied


def test_model_copy_with_update():
    instance = PersonModel(first_name="Jean", last_name="Valjean")
    copied = instance.model_copy(update={"last_name": "Pierre"})
    assert instance.id == copied.id
    assert instance.first_name == copied.first_name
    assert copied.last_name == "Pierre"


def test_model_copy_with_update_primary_key():
    instance = PersonModel(first_name="Jean", last_name="Valjean")
    copied = instance.model_copy(update={"id": ObjectId()})
    assert instance.first_name == copied.first_name
    assert copied.last_name == copied.last_name
    assert instance.id != copied.id


@pytest.mark.filterwarnings("ignore:copy is deprecated")
def test_deprecated_model_copy_call():
    class M(Model): ...

    with pytest.raises(NotImplementedError):
        M().copy(include={"id"})

    with pytest.raises(NotImplementedError):
        M().copy(exclude={"id"})


def test_model_copy_deep_embedded():
    class E(EmbeddedModel):
        f: int

    class M(Model):
        e: E

    instance = M(e=E(f=1))
    copied = instance.model_copy(deep=True)
    assert instance.e is not copied.e


def test_model_copy_deep_embedded_mutability():
    class F(EmbeddedModel):
        g: int

    class E(EmbeddedModel):
        f: F

    class M(Model):
        e: E

    instance = M(e=E(f=F(g=1)))
    copied = instance.model_copy(deep=True)
    copied.e.f.g = 42
    assert instance.e.f.g != copied.e.f.g


def test_model_copy_not_deep_embedded():
    class E(EmbeddedModel):
        f: int

    class M(Model):
        e: E

    instance = M(e=E(f=1))
    copied = instance.model_copy(deep=False)
    assert instance.e is copied.e


@pytest.mark.parametrize("deep", [True, False])
def test_model_copy_with_reference(deep: bool):
    class R(Model):
        f: int

    class M(Model):
        r: R = Reference()

    ref_instance = R(f=12)
    instance = M(r=ref_instance)
    copied = instance.model_copy(deep=deep)
    assert instance.model_dump_doc() == copied.model_dump_doc()
    assert instance.r == copied.r


@pytest.mark.parametrize("deep", [True, False])
def test_model_copy_field_modified(deep: bool):
    class M(Model):
        f: int

    instance = M(f=5)
    object.__setattr__(instance, "__fields_modified__", set())
    copied = instance.model_copy(update={"f": 12}, deep=deep)
    assert "f" in copied.__fields_modified__


@pytest.mark.parametrize("deep", [True, False])
def test_model_copy_field_modified_on_primary_field_change(deep: bool):
    class M(Model):
        f0: int
        f1: int
        f2: int

    instance = M(f0=12, f1=5, f2=6)
    object.__setattr__(instance, "__fields_modified__", set())
    copied = instance.model_copy(deep=deep)
    assert {"id", "f0", "f1", "f2"} == copied.__fields_modified__


INITIAL_FIRST_NAME, INITIAL_LAST_NAME = "INITIAL_FIRST_NAME", "INITIAL_LAST_NAME"
UPDATED_NAME = "UPDATED_NAME"


@pytest.fixture
def instance_to_update():
    return PersonModel(first_name=INITIAL_FIRST_NAME, last_name=INITIAL_LAST_NAME)


def test_update_pydantic_model(instance_to_update):
    class Update(BaseModel):
        first_name: str

    update_obj = Update(first_name=UPDATED_NAME)
    instance_to_update.model_update(update_obj)
    assert instance_to_update.first_name == UPDATED_NAME
    assert instance_to_update.last_name == INITIAL_LAST_NAME


def test_update_dictionary(instance_to_update):
    update_obj = {"first_name": UPDATED_NAME}
    instance_to_update.model_update(update_obj)
    assert instance_to_update.first_name == UPDATED_NAME
    assert instance_to_update.last_name == INITIAL_LAST_NAME


def test_update_include(instance_to_update):
    update_obj = {"first_name": UPDATED_NAME}
    instance_to_update.model_update(update_obj, include=set())
    assert instance_to_update.first_name == INITIAL_FIRST_NAME
    assert instance_to_update.last_name == INITIAL_LAST_NAME


def test_update_exclude(instance_to_update):
    update_obj = {"first_name": UPDATED_NAME}
    instance_to_update.model_update(update_obj, exclude={"first_name"})
    assert instance_to_update.first_name == INITIAL_FIRST_NAME
    assert instance_to_update.last_name == INITIAL_LAST_NAME


def test_update_exclude_none(instance_to_update):
    class Update(BaseModel):
        first_name: Optional[str]
        last_name: Optional[str]

    update_obj = Update(first_name=UPDATED_NAME, last_name=None)
    instance_to_update.model_update(update_obj, exclude_unset=False, exclude_none=True)
    assert instance_to_update.first_name == UPDATED_NAME
    assert instance_to_update.last_name == INITIAL_LAST_NAME


def test_update_exclude_defaults(instance_to_update):
    initial_instance = instance_to_update.model_copy()

    class Update(BaseModel):
        first_name: Optional[str] = None
        last_name: str = UPDATED_NAME

    update_obj = Update()
    instance_to_update.model_update(
        update_obj, exclude_unset=False, exclude_defaults=True
    )
    assert instance_to_update == initial_instance


def test_update_exclude_over_include(instance_to_update):
    update_obj = {"first_name": UPDATED_NAME}
    instance_to_update.model_update(
        update_obj, include={"first_name"}, exclude={"first_name"}
    )
    assert instance_to_update.first_name == INITIAL_FIRST_NAME
    assert instance_to_update.last_name == INITIAL_LAST_NAME


def test_update_invalid():
    class M(Model):
        f: int

    instance = M(f=12)
    update_obj = {"f": "aaa"}
    with pytest.raises(ValidationError):
        instance.model_update(update_obj)


def test_update_model_undue_update_fields():
    class M(Model):
        f: int

    instance = M(f=12)
    update_obj = {"not_in_model": "aaa"}
    instance.model_update(update_obj)


def test_update_pydantic_unset_update_fields():
    UPDATEED_VALUE = 100

    class P(BaseModel):
        f: int = UPDATEED_VALUE

    class M(Model):
        f: int

    instance = M(f=0)
    update_obj = P()
    instance.model_update(update_obj)
    assert instance.f != UPDATEED_VALUE


def test_update_pydantic_unset_update_fields_include_unset():
    UPDATEED_VALUE = 100

    class P(BaseModel):
        f: int = UPDATEED_VALUE

    class M(Model):
        f: int

    instance = M(f=0)
    update_obj = P()
    instance.model_update(update_obj, exclude_unset=False)
    assert instance.f == UPDATEED_VALUE


def test_update_embedded_model():
    class E(EmbeddedModel):
        f: int

    instance = E(f=12)
    instance.model_update({"f": 15})
    assert instance.f == 15


def test_update_reference():
    class R(Model):
        f: int

    class M(Model):
        r: R = Reference()

    r0 = R(f=0)
    r1 = R(f=1)

    instance = M(r=r0)
    instance.model_update({"r": r1})
    assert instance.r.f == r1.f
    assert instance.r == r1


def test_update_type_coercion():
    class M(Model):
        f: int

    instance = M(f=12)
    update_obj = {"f": "12"}
    instance.model_update(update_obj)
    assert isinstance(instance.f, int)


def test_update_side_effect_field_modified():
    class Rectangle(Model):
        width: float
        height: float
        area: float = 0

        @model_validator(mode="before")
        def set_area(cls, v):
            v["area"] = v["width"] * v["height"]
            return v

    r = Rectangle(width=1, height=1)
    assert r.area == 1
    r.__fields_modified__.clear()
    r.model_update({"width": 5})
    assert r.area == 5
    assert "area" in r.__fields_modified__


@pytest.mark.filterwarnings(
    "ignore: Pydantic V1 style `@root_validator` validators are deprecated"
)
def test_update_side_effect_field_modified_with_root_validator():
    class Rectangle(Model):
        width: float
        height: float
        area: float = 0

        @root_validator(skip_on_failure=True)
        def set_area(cls, v):
            v["area"] = v["width"] * v["height"]
            return v

    r = Rectangle(width=1, height=1)
    assert r.area == 1
    r.__fields_modified__.clear()
    r.model_update({"width": 5})
    assert r.area == 5
    assert "area" in r.__fields_modified__


def test_update_dict_id_exception():
    class M(Model):
        alternate_id: int = Field(primary_field=True)
        f: int

    m = M(alternate_id=0, f=0)
    with pytest.raises(ValueError, match="Updating the primary key is not supported"):
        m.model_update({"alternate_id": 1})


@pytest.mark.parametrize(
    "update_kwargs",
    (
        {"include": set()},
        {"exclude": {"alternate_id"}},
        {"include": {"alternate_id"}, "exclude": {"alternate_id"}},
    ),
)
def test_update_dict_alternate_id_filtered(update_kwargs):
    class M(Model):
        alternate_id: int = Field(primary_field=True)
        f: int

    m = M(alternate_id=0, f=0)
    m.model_update({"alternate_id": 1}, **update_kwargs)
    assert m.f == 0 and m.alternate_id == 0, "instance should be unchanged"


def test_update_pydantic_id_exception():
    class M(Model):
        alternate_id: int = Field(primary_field=True)
        f: int

    m = M(alternate_id=0, f=0)

    class UpdateObject(BaseModel):
        alternate_id: int

    with pytest.raises(ValueError, match="Updating the primary key is not supported"):
        m.model_update(UpdateObject(alternate_id=1))


@pytest.mark.parametrize(
    "update_kwargs",
    (
        {"include": set()},
        {"exclude": {"alternate_id"}},
        {"include": {"alternate_id"}, "exclude": {"alternate_id"}},
    ),
)
def test_update_pydantic_alternate_id_filtered(update_kwargs):
    class M(Model):
        alternate_id: int = Field(primary_field=True)
        f: int

    class UpdateObject(BaseModel):
        alternate_id: int

    m = M(alternate_id=0, f=0)
    m.model_update(UpdateObject(alternate_id=1), **update_kwargs)
    assert m.f == 0 and m.alternate_id == 0, "instance should be unchanged"



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Optional, Sequence, Tuple, Union, cast

import pymongo

from odmantic.field import FieldProxy
from odmantic.query import FieldProxyAny, SortExpression, asc


class ODMBaseIndex(metaclass=ABCMeta):
    def __init__(self, unique: bool, index_name: Optional[str]) -> None:
        self.unique = unique
        self.index_name = index_name

    @abstractmethod
    def get_index_specifier(self) -> Sequence[Tuple[str, int]]: ...

    def get_pymongo_index(self) -> pymongo.IndexModel:
        kwargs: Dict[str, Any] = {"keys": self.get_index_specifier()}
        if self.index_name is not None:
            kwargs["name"] = self.index_name
        if self.unique:
            kwargs["unique"] = True
        return pymongo.IndexModel(**kwargs)


class ODMSingleFieldIndex(ODMBaseIndex):
    def __init__(self, key_name: str, unique: bool, index_name: Optional[str] = None):
        super().__init__(unique, index_name)
        self.key_name = key_name

    def get_index_specifier(self) -> Sequence[Tuple[str, int]]:
        return [
            (self.key_name, pymongo.ASCENDING),
        ]


class ODMCompoundIndex(ODMBaseIndex):
    def __init__(
        self,
        fields: Tuple[SortExpression, ...],
        unique: bool,
        index_name: Optional[str],
    ):
        super().__init__(unique, index_name)
        self.fields = fields

    def get_index_specifier(self) -> Sequence[Tuple[str, int]]:
        return [
            (
                list(f.keys())[0],
                pymongo.ASCENDING if list(f.values())[0] == 1 else pymongo.DESCENDING,
            )
            for f in self.fields
        ]


class Index:
    def __init__(
        self,
        *fields: Union[FieldProxyAny, SortExpression],
        unique: bool = False,
        name: Optional[str] = None,
    ) -> None:
        """Declare an ODM index in the Model.Config.indexes generator.

        Example usage:
        ```python
        from odmantic import Model, Index
        from odmantic.query import desc

        class Player(Model):
            name: str
            score: int

            model_config = {
                "indexes": lambda: [Index(Player.name, desc(Player.score))],
            }
        ```

        Args:
            *fields (Any | SortExpression | str): fields to build the index with
            unique: build a unique index
            name: specify an optional custom index name
        """
        self.fields = cast(Tuple[Union[SortExpression, FieldProxy], ...], fields)
        self.unique = unique
        self.name = name

    def to_odm_index(self) -> "ODMBaseIndex":
        if len(self.fields) == 1:
            field = self.fields[0]
            if isinstance(field, SortExpression):
                key_name = list(field.keys())[0]
            else:
                key_name = object.__getattribute__(field, "_get_key_name")()
            return ODMSingleFieldIndex(
                key_name, unique=self.unique, index_name=self.name
            )
        else:
            fields = tuple(
                (f if isinstance(f, SortExpression) else asc(f) for f in self.fields)
            )
            return ODMCompoundIndex(fields, unique=self.unique, index_name=self.name)



from odmantic import Field, Model


class User(Model):
    name: str = Field(key_name="username")


user = User(name="John")
print(user.model_dump_doc())
#> {'username': 'John', '_id': ObjectId('5f8352a87a733b8b18b0cb27')}



from typing import List

from odmantic import Field, Model


class ExampleModel(Model):
    small_int: int = Field(le=10)
    big_int: int = Field(gt=1000)
    even_int: int = Field(multiple_of=2)

    small_float: float = Field(lt=10)
    big_float: float = Field(ge=1e10)

    short_string: str = Field(max_length=10)
    long_string: str = Field(min_length=100)
    string_starting_with_the: str = Field(regex=r"^The")

    short_str_list: List[str] = Field(max_items=5)
    long_str_list: List[str] = Field(min_items=15)



from odmantic import Model, query


class Tree(Model):
    name: str
    size: float


(Tree.name == "Spruce") & (Tree.size <= 2)
#> QueryExpression(
#>     {
#>         "$and": (
#>             QueryExpression({"name": {"$eq": "Spruce"}}),
#>             QueryExpression({"size": {"$lte": 2}}),
#>         )
#>     }
#> )
query.and_(Tree.name == "Spruce", Tree.size <= 2)
#> ... same output ...



from odmantic import AIOEngine, EmbeddedModel, Model
from odmantic.query import desc


class CapitalCity(EmbeddedModel):
    name: str
    population: int


class Country(Model):
    name: str
    currency: str
    capital_city: CapitalCity


engine = AIOEngine()
await engine.find(Country, sort=desc(Country.capital_city.population))



book = await engine.find_one(Book, Book.title == "Python Cookbook")
authors = await engine.find(Author, Author.id.in_(book.author_ids))
print(authors)
#> [
#>   Author(id=ObjectId("5f7a37dc7311be1362e1da4e"), name="David Beazley"),
#>   Author(id=ObjectId("5f7a37dc7311be1362e1da4f"), name="Brian K. Jones"),
#> ]



import datetime
import importlib.metadata
import os
from enum import Enum

import typer
from click.types import Choice
from semver import VersionInfo


class BumpType(str, Enum):
    major = "major"
    minor = "minor"
    patch = "patch"


def get_current_version() -> VersionInfo:
    version = importlib.metadata.version("odmantic")
    return VersionInfo.parse(version)


def get_new_version(current_version: VersionInfo, bump_type: BumpType) -> VersionInfo:
    if bump_type == BumpType.major:
        return current_version.bump_major()
    if bump_type == BumpType.minor:
        return current_version.bump_minor()
    if bump_type == BumpType.patch:
        return current_version.bump_patch()
    raise NotImplementedError("Unhandled bump type")


PYPROJECT_PATH = "./pyproject.toml"


def update_pyproject(current_version: VersionInfo, new_version: VersionInfo) -> None:
    with open(PYPROJECT_PATH) as f:
        content = f.read()
    new_content = content.replace(
        f'version = "{current_version}"', f'version = "{new_version}"'
    )
    if content == new_content:
        typer.secho("Couldn't bump version in pyproject.toml", fg=typer.colors.RED)
        raise typer.Exit(1)
    with open(PYPROJECT_PATH, "w") as f:
        f.write(new_content)
    typer.secho("Version updated with success", fg=typer.colors.GREEN)


RELEASE_NOTE_PATH = "./__release_notes__.md"


def get_release_notes() -> str:
    with open("./CHANGELOG.md", "r") as f:
        while not f.readline().strip() == "## [Unreleased]":
            pass
        content = ""
        while not (line := f.readline().strip()).startswith("## "):
            content = content + line + "\n"
    return content


def save_release_notes(release_notes: str) -> None:
    if os.path.exists(RELEASE_NOTE_PATH):
        typer.secho(
            f"Release note file {RELEASE_NOTE_PATH} already exists", fg=typer.colors.RED
        )
        raise typer.Exit(1)
    with open(RELEASE_NOTE_PATH, "w") as f:
        f.write(release_notes)
    typer.secho("Release note file generated with success", fg=typer.colors.GREEN)


CHANGELOG_PATH = "./CHANGELOG.md"


def update_changelog(current_version: VersionInfo, new_version: VersionInfo) -> None:
    today = datetime.date.today()
    date_str = f"{today.year}-{today.month:02d}-{today.day:02d}"
    with open(CHANGELOG_PATH, "r") as f:
        content = f.read()
    # Add version header
    content = content.replace(
        "## [Unreleased]", ("## [Unreleased]\n\n" f"## [{new_version}] - {date_str}")
    )
    # Add version links
    content = content.replace(
        f"[unreleased]: https://github.com/art049/odmantic/compare/v{current_version}...HEAD",
        (
            f"[{new_version}]: https://github.com/art049/odmantic/compare/v{current_version}...v{new_version}\n"
            f"[unreleased]: https://github.com/art049/odmantic/compare/v{new_version}...HEAD"
        ),
    )
    with open(CHANGELOG_PATH, "w") as f:
        f.write(content)
    typer.secho("Changelog updated with success", fg=typer.colors.GREEN)


VERSION_FILE_PATH = "__version__.txt"


def create_version_file(new_version: VersionInfo) -> None:
    if os.path.exists(VERSION_FILE_PATH):
        typer.secho(
            f"Version file {VERSION_FILE_PATH} already exists", fg=typer.colors.RED
        )
        raise typer.Exit(1)
    with open(VERSION_FILE_PATH, "w") as f:
        f.write(str(new_version))


def summarize(
    current_version: VersionInfo,
    new_version: VersionInfo,
    bump_type: BumpType,
    release_notes: str,
) -> None:
    typer.secho("Release summary:", fg=typer.colors.BLUE, bold=True)
    typer.secho(f"    Version bump: {bump_type.upper()}", bold=True)
    typer.secho(f"    Version change: {current_version} -> {new_version}", bold=True)
    typer.confirm("Continue to release notes preview ?", abort=True, default=True)

    release_header = typer.style(
        f"RELEASE NOTE {new_version}\n\n", fg=typer.colors.BLUE, bold=True
    )
    typer.echo_via_pager(release_header + release_notes)
    typer.confirm("Continue ?", abort=True, default=True)


def main() -> None:
    current_version = get_current_version()
    typer.secho(f"Current version: {current_version}", bold=True)
    bump_type: BumpType = typer.prompt(
        typer.style("Release type ?", fg=typer.colors.BLUE, bold=True),
        type=Choice(list(BumpType.__members__)),
        default=BumpType.patch,
        show_choices=True,
    )
    new_version = get_new_version(current_version, bump_type)
    release_notes = get_release_notes()
    summarize(current_version, new_version, bump_type, release_notes)
    save_release_notes(release_notes)
    update_pyproject(current_version, new_version)
    update_changelog(current_version, new_version)
    create_version_file(new_version)
    typer.confirm("Additionnal release commit files staged ?", abort=True, default=True)


if __name__ == "__main__":
    typer.run(main)



from odmantic import Model, query


class Tree(Model):
    name: str
    average_size: float


Tree.average_size < 2
#> QueryExpression({'average_size': {'$lt': 2}})
Tree.average_size.lt(2)
#> QueryExpression({'average_size': {'$lt': 2}})
query.lt(Tree.average_size, 2)
#> QueryExpression({'average_size': {'$lt': 2}})

Tree.average_size <= 2
#> QueryExpression({'average_size': {'$lte': 2}})
Tree.average_size.lte(2)
#> QueryExpression({'average_size': {'$lte': 2}})
query.lte(Tree.average_size, 2)
#> QueryExpression({'average_size': {'$lte': 2}})



import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient


@pytest.fixture
def fastapi_app():
    return FastAPI()


@pytest.fixture
def test_client(fastapi_app):
    return TestClient(fastapi_app)



from typing import List

from fastapi import FastAPI, HTTPException

from odmantic import AIOEngine, Model, ObjectId


class Tree(Model):
    name: str
    average_size: float
    discovery_year: int


app = FastAPI()

engine = AIOEngine()


@app.put("/trees/", response_model=Tree)
async def create_tree(tree: Tree):
    await engine.save(tree)
    return tree


@app.get("/trees/", response_model=List[Tree])
async def get_trees():
    trees = await engine.find(Tree)
    return trees


@app.get("/trees/count", response_model=int)
async def count_trees():
    count = await engine.count(Tree)
    return count


@app.get("/trees/{id}", response_model=Tree)
async def get_tree_by_id(id: ObjectId):
    tree = await engine.find_one(Tree, Tree.id == id)
    if tree is None:
        raise HTTPException(404)
    return tree



from odmantic import AIOEngine, Model


class Player(Model):
    name: str
    game: str


engine = AIOEngine()

leeroy = Player(name="Leeroy Jenkins", game="World of Warcraft")

async with engine.session() as session:
    await session.save_all(
        [
            Player(name="Shroud", game="Counter-Strike"),
            Player(name="Serral", game="Starcraft"),
            Player(name="TLO", game="Starcraft"),
        ]
    )
    player_count = await session.count(Player)
    print(player_count)
    #> 3



from odmantic import AIOEngine, Model


class Player(Model):
    name: str
    game: str


engine = AIOEngine()

delete_count = await engine.remove(
    Player, Player.game == "Warzone", just_one=True
)
print(delete_count)
#> 1



from odmantic.field import Field
from odmantic.index import Index, ODMCompoundIndex, ODMSingleFieldIndex
from odmantic.model import EmbeddedModel, Model
from odmantic.query import asc, desc


def test_single_index_definition():
    class M(Model):
        f: int = Field(index=True)

    indexes = M.__indexes__()
    assert len(indexes) == 1
    index = indexes[0]
    assert isinstance(index, ODMSingleFieldIndex)
    assert not index.unique
    assert index.key_name == "f"


def test_single_index_with_key_name_definition():
    class M(Model):
        f: int = Field(key_name="custom", index=True)

    indexes = M.__indexes__()
    assert len(indexes) == 1
    index = indexes[0]
    assert isinstance(index, ODMSingleFieldIndex)
    assert not index.unique
    assert index.key_name == "custom"


def test_single_index_unique_definition():
    class M(Model):
        f: int = Field(unique=True)

    indexes = M.__indexes__()
    assert len(indexes) == 1
    index = indexes[0]
    assert isinstance(index, ODMSingleFieldIndex)
    assert index.unique


def test_single_index_index_and_unique_definition():
    class M(Model):
        f: int = Field(index=True, unique=True)

    indexes = M.__indexes__()
    assert len(indexes) == 1
    index = indexes[0]
    assert isinstance(index, ODMSingleFieldIndex)
    assert index.unique


def test_single_index_definition_from_generator():
    class M(Model):
        f: int

        model_config = {
            "indexes": lambda: [Index(M.f, unique=True)],
        }

    indexes = M.__indexes__()
    assert len(indexes) == 1
    index = indexes[0]
    assert isinstance(index, ODMSingleFieldIndex)
    assert index.unique


def test_compound_index_definition():
    class M(Model):
        f: int
        g: str

        model_config = {
            "indexes": lambda: [Index(M.f, desc(M.g), unique=True)],
        }

    indexes = M.__indexes__()
    assert len(indexes) == 1
    index = indexes[0]
    assert isinstance(index, ODMCompoundIndex)
    assert index.fields == (asc(M.f), desc(M.g))
    assert index.unique


def test_multiple_indexes_definition():
    class M(Model):
        f: int
        g: str
        h: float = Field(index=True)

        model_config = {
            "indexes": lambda: [
                Index(M.f, desc(M.g), unique=True, name="asc_desc"),
                Index(M.g),
            ],
        }

    indexes = M.__indexes__()
    assert len(indexes) == 3

    assert isinstance(indexes[0], ODMSingleFieldIndex)
    assert indexes[0].key_name == "h"

    assert isinstance(indexes[1], ODMCompoundIndex)
    assert indexes[1].fields == (asc(M.f), desc(M.g))
    assert indexes[1].unique
    assert indexes[1].index_name == "asc_desc"
    assert indexes[1].unique

    assert isinstance(indexes[2], ODMSingleFieldIndex)
    assert not indexes[2].unique


def test_embedded_index_definition():
    class E(EmbeddedModel):
        f: int

    class M(Model):
        e: E = Field(index=True)

    indexes = M.__indexes__()
    assert len(indexes) == 1
    index = indexes[0]

    assert isinstance(index, ODMSingleFieldIndex)
    assert index.key_name == "e"


def test_embedded_index_definition_generator():
    class E(EmbeddedModel):
        f: int

    class M(Model):
        e: E

        model_config = {
            "indexes": lambda: [Index(M.e)],
        }

    indexes = M.__indexes__()
    assert len(indexes) == 1
    index = indexes[0]

    assert isinstance(index, ODMSingleFieldIndex)
    assert index.key_name == "e"


def test_embedded_field_index_definition():
    class E(EmbeddedModel):
        f: int

    class M(Model):
        e: E

        model_config = {
            "indexes": lambda: [Index(M.e.f)],
        }

    indexes = M.__indexes__()
    assert len(indexes) == 1
    index = indexes[0]

    assert isinstance(index, ODMSingleFieldIndex)
    assert index.key_name == "e.f"



from odmantic.query import QueryExpression, SortExpression, asc
from tests.zoo.book_embedded import Book, Publisher
from tests.zoo.tree import TreeKind, TreeModel


def test_embedded_eq():
    pub = Publisher(name="O'Reilly Media", founded=1980, location="CA")
    assert (Book.publisher == pub) == {
        "publisher": {
            "$eq": {"name": "O'Reilly Media", "founded": 1980, "location": "CA"}
        }
    }


def test_embedded_eq_on_subfield():
    assert (Book.publisher.location == "EU") == {"publisher.location": {"$eq": "EU"}}


def test_eq_on_enum():
    assert (TreeModel.kind == TreeKind.BIG) == {"kind": {"$eq": TreeKind.BIG.value}}
    assert (TreeModel.kind == "big") == {"kind": {"$eq": "big"}}


def test_query_repr():
    assert (
        repr(TreeModel.name == "Spruce")
        == "QueryExpression({'name': {'$eq': 'Spruce'}})"
    )


def test_query_empty_repr():
    assert repr(QueryExpression()) == "QueryExpression()"


def test_sort_repr():
    assert repr(asc(TreeModel.name)) == "SortExpression({'name': 1})"


def test_sort_empty_repr():
    assert repr(SortExpression()) == "SortExpression()"



from typing import Union

from odmantic import Model


class Thing(Model):
    ref_id: Union[int, str]


thing_1 = Thing(ref_id=42)
print(thing_1.ref_id)
#> 42

thing_2 = Thing(ref_id="i am a string")
print(thing_2.ref_id)
#> i am a string



from odmantic import AIOEngine, Model


class Rectangle(Model):
    length: float
    width: float


rectangles = [
    Rectangle(length=0.1, width=1),
    Rectangle(length=3.5, width=1),
    Rectangle(length=2.87, width=5.19),
    Rectangle(length=1, width=10),
    Rectangle(length=0.1, width=100),
]

engine = AIOEngine()
await engine.save_all(rectangles)

collection = engine.get_collection(Rectangle)
pipeline = []
# Add an area field
pipeline.append(
    {
        "$addFields": {
            "area": {
                "$multiply": [++Rectangle.length, ++Rectangle.width]
            }  # Compute the area remotely
        }
    }
)
# Filter only rectanges with an area lower than 10
pipeline.append({"$match": {"area": {"$lt": 10}}})
# Project to keep only the defined fields (this step is optional)
pipeline.append(
    {
        "$project": {
            +Rectangle.length: True,
            +Rectangle.width: True,
        }  # Specifying "area": False is unnecessary here
    }
)
documents = await collection.aggregate(pipeline).to_list(length=None)
small_rectangles = [Rectangle.model_validate_doc(doc) for doc in documents]
print(small_rectangles)
#> [
#>     Rectangle(id=ObjectId("..."), length=0.1, width=1.0),
#>     Rectangle(id=ObjectId("..."), length=3.5, width=1.0),
#> ]



from odmantic import Model, SyncEngine


class Player(Model):
    name: str
    game: str


engine = SyncEngine()

player_count = engine.count(Player)
print(player_count)
#> 4
cs_count = engine.count(Player, Player.game == "Counter-Strike")
print(cs_count)
#> 1
valorant_count = engine.count(Player, Player.game == "Valorant")
print(valorant_count)
#> 0



from odmantic import Model, query


class Tree(Model):
    name: str
    average_size: float


Tree.name.in_(["Spruce", "Pine"])
#> QueryExpression({'name': {'$in': ['Spruce', 'Pine']}})
query.in_(Tree.name, ["Spruce", "Pine"])
#> QueryExpression({'name': {'$in': ['Spruce', 'Pine']}})



from odmantic import SyncEngine, Model


class Player(Model):
    name: str
    game: str


engine = SyncEngine()

delete_count = engine.remove(Player, Player.game == "Warzone")
print(delete_count)
#> 2



from typing import List

from bson import ObjectId

from odmantic import Model


class TwitterUser(Model):
    """Self referencing model with manual reference handling"""

    following: List[ObjectId] = []



from bson import ObjectId

from odmantic import Field, Model


class User(Model):
    name: str = Field(key_name="username")


document = {"username": "John", "_id": ObjectId("5f8352a87a733b8b18b0cb27")}

user = User.model_validate_doc(document)
print(repr(user))
#> User(id=ObjectId('5f8352a87a733b8b18b0cb27'), name='John')



import pytest

from odmantic.engine import AIOEngine, SyncEngine
from odmantic.session import AIOTransaction, SyncTransaction
from tests.integration.conftest import only_on_replica

from ..zoo.person import PersonModel

pytestmark = pytest.mark.asyncio


class CustomException(Exception):
    pass


async def test_session_exception_propagation(aio_engine: AIOEngine):
    with pytest.raises(CustomException):
        async with aio_engine.session():
            raise CustomException()


def test_sync_session_exception_propagation(sync_engine: SyncEngine):
    with pytest.raises(CustomException):
        with sync_engine.session():
            raise CustomException()


@only_on_replica
async def test_transaction_exception_propagation(aio_engine: AIOEngine):
    with pytest.raises(CustomException):
        async with aio_engine.transaction():
            raise CustomException()


@only_on_replica
def test_sync_transaction_exception_propagation(sync_engine: SyncEngine):
    with pytest.raises(CustomException):
        with sync_engine.session():
            raise CustomException()


async def test_start_session_twice(aio_engine: AIOEngine):
    with pytest.raises(RuntimeError, match="Session is already started"):
        async with aio_engine.session() as session:
            await session.start()


def test_sync_start_session_twice(sync_engine: SyncEngine):
    with pytest.raises(RuntimeError, match="Session is already started"):
        with sync_engine.session() as session:
            session.start()


async def test_end_a_non_started_session(aio_engine: AIOEngine):
    with pytest.raises(RuntimeError, match="Session is not started"):
        await aio_engine.session().end()


def test_sync_end_a_non_started_session(sync_engine: SyncEngine):
    with pytest.raises(RuntimeError, match="Session is not started"):
        sync_engine.session().end()


@only_on_replica
async def test_start_transaction_twice(aio_engine: AIOEngine):
    with pytest.raises(RuntimeError, match="Transaction already started"):
        async with aio_engine.transaction() as transaction:
            await transaction.start()


@only_on_replica
def test_sync_start_transaction_twice(sync_engine: SyncEngine):
    with pytest.raises(RuntimeError, match="Transaction already started"):
        with sync_engine.transaction() as transaction:
            transaction.start()


@only_on_replica
async def test_abort_a_non_started_transaction(aio_engine: AIOEngine):
    with pytest.raises(RuntimeError, match="Transaction not started"):
        await aio_engine.transaction().abort()


@only_on_replica
def test_sync_abort_a_non_started_transaction(sync_engine: SyncEngine):
    with pytest.raises(RuntimeError, match="Transaction not started"):
        sync_engine.transaction().abort()


@only_on_replica
async def test_commit_a_non_started_transaction(aio_engine: AIOEngine):
    with pytest.raises(RuntimeError, match="Transaction not started"):
        await aio_engine.transaction().commit()


@only_on_replica
def test_sync_commit_a_non_started_transaction(sync_engine: SyncEngine):
    with pytest.raises(RuntimeError, match="Transaction not started"):
        sync_engine.transaction().commit()


@only_on_replica
async def test_create_transaction_with_a_non_started_session(aio_engine: AIOEngine):
    with pytest.raises(RuntimeError, match="provided session is not started"):
        session = aio_engine.session()
        AIOTransaction(session)


@only_on_replica
async def test_sync_create_transaction_with_a_non_started_session(
    sync_engine: SyncEngine,
):
    with pytest.raises(RuntimeError, match="provided session is not started"):
        session = sync_engine.session()
        SyncTransaction(session)


async def test_operation_on_ended_session_should_fail(aio_engine: AIOEngine):
    session = aio_engine.session()
    await session.start()
    await session.end()
    with pytest.raises(RuntimeError, match="session not started"):
        await session.save(PersonModel(first_name="Jean-Pierre", last_name="Pernaud"))


def test_sync_operation_on_ended_session_should_fail(sync_engine: SyncEngine):
    session = sync_engine.session()
    session.start()
    session.end()
    with pytest.raises(RuntimeError, match="session not started"):
        session.save(PersonModel(first_name="Jean-Pierre", last_name="Pernaud"))


@only_on_replica
async def test_operation_on_aborted_transaction_should_fail(aio_engine: AIOEngine):
    transaction = aio_engine.transaction()
    await transaction.start()
    await transaction.abort()
    with pytest.raises(RuntimeError, match="transaction not started"):
        await transaction.save(
            PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
        )


@only_on_replica
def test_sync_operation_on_aborted_transaction_should_fail(sync_engine: SyncEngine):
    transaction = sync_engine.transaction()
    transaction.start()
    transaction.abort()
    with pytest.raises(RuntimeError, match="transaction not started"):
        transaction.save(PersonModel(first_name="Jean-Pierre", last_name="Pernaud"))


@only_on_replica
async def test_operation_on_comitted_transaction_should_fail(aio_engine: AIOEngine):
    transaction = aio_engine.transaction()
    await transaction.start()
    await transaction.commit()
    with pytest.raises(RuntimeError, match="transaction not started"):
        await transaction.save(
            PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
        )


@only_on_replica
def test_sync_operation_on_comitted_transaction_should_fail(sync_engine: SyncEngine):
    transaction = sync_engine.transaction()
    transaction.start()
    transaction.commit()
    with pytest.raises(RuntimeError, match="transaction not started"):
        transaction.save(PersonModel(first_name="Jean-Pierre", last_name="Pernaud"))


async def test_operation_on_exited_context_session_sould_fail(aio_engine: AIOEngine):
    async with aio_engine.session() as session:
        pass
    with pytest.raises(RuntimeError, match="session not started"):
        await session.save(PersonModel(first_name="Jean-Pierre", last_name="Pernaud"))


async def test_sync_operation_on_exited_context_session_sould_fail(
    sync_engine: SyncEngine,
):
    with sync_engine.session() as session:
        pass
    with pytest.raises(RuntimeError, match="session not started"):
        session.save(PersonModel(first_name="Jean-Pierre", last_name="Pernaud"))


@only_on_replica
async def test_operation_on_exited_context_transaction_sould_fail(
    aio_engine: AIOEngine,
):
    async with aio_engine.transaction() as transaction:
        pass
    with pytest.raises(RuntimeError, match="transaction not started"):
        await transaction.save(
            PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
        )


@only_on_replica
async def test_sync_operation_on_exited_context_transaction_sould_fail(
    sync_engine: SyncEngine,
):
    with sync_engine.transaction() as transaction:
        pass
    with pytest.raises(RuntimeError, match="transaction not started"):
        transaction.save(PersonModel(first_name="Jean-Pierre", last_name="Pernaud"))


@only_on_replica
async def test_operation_on_exited_context_aborted_transaction_sould_fail(
    aio_engine: AIOEngine,
):
    async with aio_engine.transaction() as transaction:
        await transaction.abort()
    with pytest.raises(RuntimeError, match="transaction not started"):
        await transaction.find_one(PersonModel, PersonModel.first_name == "Jean-Pierre")


@only_on_replica
def test_sync_operation_on_exited_context_aborted_transaction_sould_fail(
    sync_engine: SyncEngine,
):
    with sync_engine.transaction() as transaction:
        transaction.abort()
    with pytest.raises(RuntimeError, match="transaction not started"):
        transaction.find_one(PersonModel, PersonModel.first_name == "Jean-Pierre")


@only_on_replica
async def test_session_stopped_on_manual_transaction_commit(
    aio_engine: AIOEngine,
):
    transaction = aio_engine.transaction()
    await transaction.start()
    await transaction.commit()
    assert not transaction.session.is_started


@only_on_replica
def test_sync_session_stopped_on_manual_transaction_commit(
    sync_engine: SyncEngine,
):
    transaction = sync_engine.transaction()
    transaction.start()
    transaction.commit()
    assert not transaction.session.is_started


@only_on_replica
async def test_session_stopped_on_manual_transaction_abort(
    aio_engine: AIOEngine,
):
    transaction = aio_engine.transaction()
    await transaction.start()
    await transaction.abort()
    assert not transaction.session.is_started


@only_on_replica
def test_sync_session_stopped_on_manual_transaction_abort(
    sync_engine: SyncEngine,
):
    transaction = sync_engine.transaction()
    transaction.start()
    transaction.abort()
    assert not transaction.session.is_started


@only_on_replica
async def test_abort_transaction_keep_provided_session_opened(aio_engine: AIOEngine):
    async with aio_engine.session() as session:
        async with session.transaction() as transaction:
            await transaction.abort()
        assert session.is_started


@only_on_replica
def test_sync_abort_transaction_keep_provided_session_opened(sync_engine: SyncEngine):
    with sync_engine.session() as session:
        with session.transaction() as transaction:
            transaction.abort()
        assert session.is_started


async def test_save_find_find_one_session(aio_engine: AIOEngine):
    initial_instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    async with aio_engine.session() as session:
        await session.save(initial_instance)
        found_instances = await session.find(PersonModel)
        assert len(found_instances) == 1
        assert found_instances[0].first_name == initial_instance.first_name
        assert found_instances[0].last_name == initial_instance.last_name

        single_fetched_instance = await session.find_one(
            PersonModel, PersonModel.first_name == "Jean-Pierre"
        )
        assert single_fetched_instance is not None
        assert single_fetched_instance.first_name == initial_instance.first_name
        assert single_fetched_instance.last_name == initial_instance.last_name


def test_sync_save_find_find_one_session(sync_engine: SyncEngine):
    initial_instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    with sync_engine.session() as session:
        session.save(initial_instance)
        found_instances = list(session.find(PersonModel))
        assert len(found_instances) == 1
        assert found_instances[0].first_name == initial_instance.first_name
        assert found_instances[0].last_name == initial_instance.last_name

        single_fetched_instance = session.find_one(
            PersonModel, PersonModel.first_name == "Jean-Pierre"
        )
        assert single_fetched_instance is not None
        assert single_fetched_instance.first_name == initial_instance.first_name
        assert single_fetched_instance.last_name == initial_instance.last_name


@only_on_replica
async def test_save_find_find_one_session_transaction(aio_engine: AIOEngine):
    initial_instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    async with aio_engine.session() as session:
        async with session.transaction() as tx:
            await tx.save(initial_instance)
            found_instances = await tx.find(PersonModel)
            assert len(found_instances) == 1
            assert found_instances[0].first_name == initial_instance.first_name
            assert found_instances[0].last_name == initial_instance.last_name

            single_fetched_instance = await tx.find_one(
                PersonModel, PersonModel.first_name == "Jean-Pierre"
            )
            assert single_fetched_instance is not None
            assert single_fetched_instance.first_name == initial_instance.first_name
            assert single_fetched_instance.last_name == initial_instance.last_name
            await tx.commit()


@only_on_replica
def test_sync_save_find_find_one_session_transaction(sync_engine: SyncEngine):
    initial_instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    with sync_engine.session() as session:
        with session.transaction() as tx:
            tx.save(initial_instance)
            found_instances = list(tx.find(PersonModel))
            assert len(found_instances) == 1
            assert found_instances[0].first_name == initial_instance.first_name
            assert found_instances[0].last_name == initial_instance.last_name

            single_fetched_instance = tx.find_one(
                PersonModel, PersonModel.first_name == "Jean-Pierre"
            )
            assert single_fetched_instance is not None
            assert single_fetched_instance.first_name == initial_instance.first_name
            assert single_fetched_instance.last_name == initial_instance.last_name
            tx.commit()


@only_on_replica
async def test_save_transaction_abort(aio_engine: AIOEngine):
    initial_instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    async with aio_engine.transaction() as tx:
        await tx.save(initial_instance)
        found_instances = await tx.find(PersonModel)
        assert len(found_instances) == 1
        assert found_instances[0].first_name == initial_instance.first_name
        assert found_instances[0].last_name == initial_instance.last_name
        await tx.abort()

    single_fetched_instance = await aio_engine.find_one(
        PersonModel, PersonModel.first_name == "Jean-Pierre"
    )
    assert single_fetched_instance is None


@only_on_replica
def test_sync_save_transaction_abort(sync_engine: SyncEngine):
    initial_instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    with sync_engine.transaction() as tx:
        tx.save(initial_instance)
        found_instances = list(tx.find(PersonModel))
        assert len(found_instances) == 1
        assert found_instances[0].first_name == initial_instance.first_name
        assert found_instances[0].last_name == initial_instance.last_name
        tx.abort()
    single_fetched_instance = sync_engine.find_one(
        PersonModel, PersonModel.first_name == "Jean-Pierre"
    )
    assert single_fetched_instance is None



from bson import ObjectId

from odmantic import Model


class Player(Model):
    name: str
    level: int = 1


document = {"name": "Leeroy", "_id": ObjectId("5f8352a87a733b8b18b0cb27")}

user = Player.model_validate_doc(document)
print(repr(user))
#> Player(
#>     id=ObjectId("5f8352a87a733b8b18b0cb27"),
#>     name="Leeroy",
#>     level=1,
#> )



from typing import Dict, List, Optional, Tuple

import pytest
from bson import ObjectId as BsonObjectId
from inline_snapshot import snapshot
from motor.motor_asyncio import AsyncIOMotorClient
from pymongo import MongoClient

from odmantic.bson import ObjectId
from odmantic.engine import AIOEngine, SyncEngine
from odmantic.exceptions import DocumentNotFoundError, DocumentParsingError
from odmantic.field import Field
from odmantic.model import EmbeddedModel, Model
from odmantic.query import asc, desc
from tests.integration.conftest import only_on_replica
from tests.integration.utils import redact_objectid
from tests.zoo.book_reference import Book, Publisher

from ..zoo.person import PersonModel

pytestmark = pytest.mark.asyncio


def test_default_motor_client_creation():
    engine = AIOEngine()
    assert isinstance(engine.client, AsyncIOMotorClient)


def test_no_motor_raises_for_aioengine_client_creation():
    import odmantic.engine

    motor = odmantic.engine.motor
    odmantic.engine.motor = None
    with pytest.raises(RuntimeError) as e:
        AIOEngine()
    assert 'pip install "odmantic[motor]"' in str(e)
    odmantic.engine.motor = motor


def test_default_pymongo_client_creation():
    engine = SyncEngine()
    assert isinstance(engine.client, MongoClient)


def test_no_motor_passes_with_syncengine_client_creation():
    import odmantic.engine

    motor = odmantic.engine.motor
    odmantic.engine.motor = None
    engine = SyncEngine()
    assert isinstance(engine.client, MongoClient)
    odmantic.engine.motor = motor


@pytest.mark.parametrize("illegal_character", ("/", "\\", ".", '"', "$"))
def test_invalid_database_name(illegal_character: str):
    with pytest.raises(ValueError, match="database name cannot contain"):
        AIOEngine(database=f"prefix{illegal_character}suffix")
    with pytest.raises(ValueError, match="database name cannot contain"):
        SyncEngine(database=f"prefix{illegal_character}suffix")


async def test_save(aio_engine: AIOEngine):
    instance = await aio_engine.save(
        PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    )
    assert isinstance(instance.id, BsonObjectId)


def test_sync_save(sync_engine: SyncEngine):
    instance = sync_engine.save(
        PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    )
    assert isinstance(instance.id, BsonObjectId)


async def test_save_find_find_one(aio_engine: AIOEngine):
    initial_instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    await aio_engine.save(initial_instance)
    found_instances = await aio_engine.find(PersonModel)
    assert len(found_instances) == 1
    assert found_instances[0].first_name == initial_instance.first_name
    assert found_instances[0].last_name == initial_instance.last_name

    single_fetched_instance = await aio_engine.find_one(
        PersonModel, PersonModel.first_name == "Jean-Pierre"
    )
    assert single_fetched_instance is not None
    assert single_fetched_instance.first_name == initial_instance.first_name
    assert single_fetched_instance.last_name == initial_instance.last_name


def test_sync_save_find_find_one(sync_engine: SyncEngine):
    initial_instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    sync_engine.save(initial_instance)
    found_instances = list(sync_engine.find(PersonModel))
    assert len(found_instances) == 1
    assert found_instances[0].first_name == initial_instance.first_name
    assert found_instances[0].last_name == initial_instance.last_name

    single_fetched_instance = sync_engine.find_one(
        PersonModel, PersonModel.first_name == "Jean-Pierre"
    )
    assert single_fetched_instance is not None
    assert single_fetched_instance.first_name == initial_instance.first_name
    assert single_fetched_instance.last_name == initial_instance.last_name


async def test_find_one_not_existing(aio_engine: AIOEngine):
    fetched = await aio_engine.find_one(PersonModel)
    assert fetched is None


def test_sync_find_one_not_existing(sync_engine: SyncEngine):
    fetched = sync_engine.find_one(PersonModel)
    assert fetched is None


@pytest.fixture(scope="function")
async def person_persisted(aio_engine: AIOEngine):
    initial_instances = [
        PersonModel(first_name="Jean-Pierre", last_name="Pernaud"),
        PersonModel(first_name="Jean-Pierre", last_name="Castaldi"),
        PersonModel(first_name="Michel", last_name="Drucker"),
    ]
    return await aio_engine.save_all(initial_instances)


async def test_save_multiple_simple_find_find_one(
    aio_engine: AIOEngine, person_persisted: List[PersonModel]
):
    found_instances = await aio_engine.find(
        PersonModel, PersonModel.first_name == "Michel"
    )
    assert len(found_instances) == 1
    assert found_instances[0].first_name == person_persisted[2].first_name
    assert found_instances[0].last_name == person_persisted[2].last_name

    found_instances = await aio_engine.find(
        PersonModel, PersonModel.first_name == "Jean-Pierre"
    )
    assert len(found_instances) == 2
    assert found_instances[0].id != found_instances[1].id

    single_retrieved = await aio_engine.find_one(
        PersonModel, PersonModel.first_name == "Jean-Pierre"
    )

    assert single_retrieved is not None
    assert single_retrieved in person_persisted


def test_sync_save_multiple_simple_find_find_one(
    sync_engine: SyncEngine, person_persisted: List[PersonModel]
):
    found_instances = list(
        sync_engine.find(PersonModel, PersonModel.first_name == "Michel")
    )
    assert len(found_instances) == 1
    assert found_instances[0].first_name == person_persisted[2].first_name
    assert found_instances[0].last_name == person_persisted[2].last_name

    found_instances = list(
        sync_engine.find(PersonModel, PersonModel.first_name == "Jean-Pierre")
    )
    assert len(found_instances) == 2
    assert found_instances[0].id != found_instances[1].id

    single_retrieved = sync_engine.find_one(
        PersonModel, PersonModel.first_name == "Jean-Pierre"
    )

    assert single_retrieved is not None
    assert single_retrieved in person_persisted


async def test_find_sync_iteration(
    aio_engine: AIOEngine, person_persisted: List[PersonModel]
):
    fetched = set()
    for inst in await aio_engine.find(PersonModel):
        fetched.add(inst.id)

    assert set(i.id for i in person_persisted) == fetched


def test_sync_find_sync_iteration(
    sync_engine: SyncEngine, person_persisted: List[PersonModel]
):
    fetched = set()
    for inst in sync_engine.find(PersonModel):
        fetched.add(inst.id)

    assert set(i.id for i in person_persisted) == fetched


@pytest.mark.usefixtures("person_persisted")
async def test_find_sync_iteration_cached(aio_engine: AIOEngine, aio_mock_collection):
    cursor = aio_engine.find(PersonModel)
    initial = await cursor
    collection = aio_mock_collection()
    cached = await cursor
    collection.aggregate.assert_not_awaited()
    assert cached == initial


@pytest.mark.usefixtures("person_persisted")
def test_sync_find_sync_iteration_cached(sync_engine: SyncEngine, sync_mock_collection):
    cursor = sync_engine.find(PersonModel)
    initial = list(cursor)
    collection = sync_mock_collection()
    cached = list(cursor)
    collection.aggregate.assert_not_called()
    assert cached == initial


@pytest.mark.usefixtures("person_persisted")
async def test_find_async_iteration_cached(aio_engine: AIOEngine, aio_mock_collection):
    cursor = aio_engine.find(PersonModel)
    initial = []
    async for inst in cursor:
        initial.append(inst)
    collection = aio_mock_collection()
    cached = []
    async for inst in cursor:
        cached.append(inst)
    collection.aggregate.assert_not_awaited()
    assert cached == initial


@pytest.mark.usefixtures("person_persisted")
def test_sync_find_async_iteration_cached(
    sync_engine: SyncEngine, sync_mock_collection
):
    cursor = sync_engine.find(PersonModel)
    initial = []
    for inst in cursor:
        initial.append(inst)
    collection = sync_mock_collection()
    cached = []
    for inst in cursor:
        cached.append(inst)
    collection.aggregate.assert_not_called()
    assert cached == initial


async def test_find_skip(aio_engine: AIOEngine, person_persisted: List[PersonModel]):
    results = await aio_engine.find(PersonModel, skip=1)
    assert len(results) == 2
    for instance in results:
        assert instance in person_persisted


def test_sync_find_skip(sync_engine: SyncEngine, person_persisted: List[PersonModel]):
    results = list(sync_engine.find(PersonModel, skip=1))
    assert len(results) == 2
    for instance in results:
        assert instance in person_persisted


async def test_find_one_bad_query(aio_engine: AIOEngine):
    with pytest.raises(TypeError):
        await aio_engine.find_one(PersonModel, True, False)


def test_sync_find_one_bad_query(sync_engine: SyncEngine):
    with pytest.raises(TypeError):
        sync_engine.find_one(PersonModel, True, False)


async def test_find_one_on_non_model(aio_engine: AIOEngine):
    class BadModel:
        pass

    with pytest.raises(TypeError):
        await aio_engine.find_one(BadModel)  # type: ignore


def test_sync_find_one_on_non_model(sync_engine: SyncEngine):
    class BadModel:
        pass

    with pytest.raises(TypeError):
        sync_engine.find_one(BadModel)  # type: ignore


async def test_find_invalid_limit(aio_engine: AIOEngine):
    with pytest.raises(ValueError):
        await aio_engine.find(PersonModel, limit=0)
    with pytest.raises(ValueError):
        await aio_engine.find(PersonModel, limit=-12)


def test_sync_find_invalid_limit(sync_engine: SyncEngine):
    with pytest.raises(ValueError):
        sync_engine.find(PersonModel, limit=0)
    with pytest.raises(ValueError):
        sync_engine.find(PersonModel, limit=-12)


async def test_find_invalid_skip(aio_engine: AIOEngine):
    with pytest.raises(ValueError):
        await aio_engine.find(PersonModel, skip=-1)
    with pytest.raises(ValueError):
        await aio_engine.find(PersonModel, limit=-12)


def test_sync_find_invalid_skip(sync_engine: SyncEngine):
    with pytest.raises(ValueError):
        sync_engine.find(PersonModel, skip=-1)
    with pytest.raises(ValueError):
        sync_engine.find(PersonModel, limit=-12)


@pytest.mark.usefixtures("person_persisted")
async def test_skip(aio_engine: AIOEngine):
    p = await aio_engine.find(PersonModel, skip=1)
    assert len(p) == 2


@pytest.mark.usefixtures("person_persisted")
def test_sync_skip(sync_engine: SyncEngine):
    p = list(sync_engine.find(PersonModel, skip=1))
    assert len(p) == 2


@pytest.mark.usefixtures("person_persisted")
async def test_limit(aio_engine: AIOEngine):
    p = await aio_engine.find(PersonModel, limit=1)
    assert len(p) == 1


@pytest.mark.usefixtures("person_persisted")
def test_sync_limit(sync_engine: SyncEngine):
    p = list(sync_engine.find(PersonModel, limit=1))
    assert len(p) == 1


@pytest.mark.usefixtures("person_persisted")
async def test_skip_limit(aio_engine: AIOEngine):
    p = await aio_engine.find(PersonModel, skip=1, limit=1)
    assert len(p) == 1


@pytest.mark.usefixtures("person_persisted")
def test_sync_skip_limit(sync_engine: SyncEngine):
    p = list(sync_engine.find(PersonModel, skip=1, limit=1))
    assert len(p) == 1


async def test_save_multiple_time_same_document(aio_engine: AIOEngine):
    fixed_id = ObjectId()

    instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud", id=fixed_id)
    await aio_engine.save(instance)

    instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud", id=fixed_id)
    await aio_engine.save(instance)

    assert await aio_engine.count(PersonModel, PersonModel.id == fixed_id) == 1


def test_sync_save_multiple_time_same_document(sync_engine: SyncEngine):
    fixed_id = ObjectId()

    instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud", id=fixed_id)
    sync_engine.save(instance)

    instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud", id=fixed_id)
    sync_engine.save(instance)

    assert sync_engine.count(PersonModel, PersonModel.id == fixed_id) == 1


@pytest.mark.usefixtures("person_persisted")
async def test_count(aio_engine: AIOEngine):
    count = await aio_engine.count(PersonModel)
    assert count == 3

    count = await aio_engine.count(PersonModel, PersonModel.first_name == "Michel")
    assert count == 1

    count = await aio_engine.count(PersonModel, PersonModel.first_name == "Gérard")
    assert count == 0


@pytest.mark.usefixtures("person_persisted")
def test_sync_count(sync_engine: SyncEngine):
    count = sync_engine.count(PersonModel)
    assert count == 3

    count = sync_engine.count(PersonModel, PersonModel.first_name == "Michel")
    assert count == 1

    count = sync_engine.count(PersonModel, PersonModel.first_name == "Gérard")
    assert count == 0


async def test_count_on_non_model_fails(aio_engine: AIOEngine):
    class BadModel:
        pass

    with pytest.raises(TypeError):
        await aio_engine.count(BadModel)  # type: ignore


def test_sync_count_on_non_model_fails(sync_engine: SyncEngine):
    class BadModel:
        pass

    with pytest.raises(TypeError):
        sync_engine.count(BadModel)  # type: ignore


async def test_find_on_embedded_raises(aio_engine: AIOEngine):
    class BadModel(EmbeddedModel):
        field: int

    with pytest.raises(TypeError):
        await aio_engine.find(BadModel)  # type: ignore


def test_sync_find_on_embedded_raises(sync_engine: SyncEngine):
    class BadModel(EmbeddedModel):
        field: int

    with pytest.raises(TypeError):
        sync_engine.find(BadModel)  # type: ignore


async def test_save_on_embedded(aio_engine: AIOEngine):
    class BadModel(EmbeddedModel):
        field: int

    instance = BadModel(field=12)
    with pytest.raises(TypeError):
        await aio_engine.save(instance)  # type: ignore


def test_sync_save_on_embedded(sync_engine: SyncEngine):
    class BadModel(EmbeddedModel):
        field: int

    instance = BadModel(field=12)
    with pytest.raises(TypeError):
        sync_engine.save(instance)  # type: ignore


@pytest.mark.usefixtures("person_persisted")
async def test_implicit_and(aio_engine: AIOEngine):
    count = await aio_engine.count(
        PersonModel,
        PersonModel.first_name == "Michel",
        PersonModel.last_name == "Drucker",
    )
    assert count == 1


@pytest.mark.usefixtures("person_persisted")
def test_sync_implicit_and(sync_engine: SyncEngine):
    count = sync_engine.count(
        PersonModel,
        PersonModel.first_name == "Michel",
        PersonModel.last_name == "Drucker",
    )
    assert count == 1


async def test_save_update(aio_engine: AIOEngine):
    instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    await aio_engine.save(instance)
    assert await aio_engine.count(PersonModel, PersonModel.last_name == "Pernaud") == 1
    instance.last_name = "Dupuis"
    await aio_engine.save(instance)
    assert await aio_engine.count(PersonModel, PersonModel.last_name == "Pernaud") == 0
    assert await aio_engine.count(PersonModel, PersonModel.last_name == "Dupuis") == 1


def test_sync_save_update(sync_engine: SyncEngine):
    instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    sync_engine.save(instance)
    assert sync_engine.count(PersonModel, PersonModel.last_name == "Pernaud") == 1
    instance.last_name = "Dupuis"
    sync_engine.save(instance)
    assert sync_engine.count(PersonModel, PersonModel.last_name == "Pernaud") == 0
    assert sync_engine.count(PersonModel, PersonModel.last_name == "Dupuis") == 1


async def test_delete_and_count(
    aio_engine: AIOEngine, person_persisted: List[PersonModel]
):
    await aio_engine.delete(person_persisted[0])
    assert await aio_engine.count(PersonModel) == 2
    await aio_engine.delete(person_persisted[1])
    assert await aio_engine.count(PersonModel) == 1
    await aio_engine.delete(person_persisted[2])
    assert await aio_engine.count(PersonModel) == 0


def test_sync_delete_and_count(
    sync_engine: SyncEngine, person_persisted: List[PersonModel]
):
    sync_engine.delete(person_persisted[0])
    assert sync_engine.count(PersonModel) == 2
    sync_engine.delete(person_persisted[1])
    assert sync_engine.count(PersonModel) == 1
    sync_engine.delete(person_persisted[2])
    assert sync_engine.count(PersonModel) == 0


async def test_delete_not_existing(aio_engine: AIOEngine):
    non_persisted_instance = PersonModel(first_name="Jean", last_name="Paul")
    with pytest.raises(DocumentNotFoundError) as exc:
        await aio_engine.delete(non_persisted_instance)
    assert exc.value.instance == non_persisted_instance


def test_sync_delete_not_existing(sync_engine: SyncEngine):
    non_persisted_instance = PersonModel(first_name="Jean", last_name="Paul")
    with pytest.raises(DocumentNotFoundError) as exc:
        sync_engine.delete(non_persisted_instance)
    assert exc.value.instance == non_persisted_instance


@pytest.mark.usefixtures("person_persisted")
async def test_remove_and_count(aio_engine: AIOEngine):
    actual_delete_count = await aio_engine.remove(
        PersonModel, PersonModel.first_name == "Jean-Pierre"
    )
    assert actual_delete_count == 2
    assert await aio_engine.count(PersonModel) == 1


@pytest.mark.usefixtures("person_persisted")
def test_sync_remove_and_count(sync_engine: SyncEngine):
    actual_delete_count = sync_engine.remove(
        PersonModel, PersonModel.first_name == "Jean-Pierre"
    )
    assert actual_delete_count == 2
    assert sync_engine.count(PersonModel) == 1


@pytest.mark.usefixtures("person_persisted")
async def test_remove_just_one(aio_engine: AIOEngine):
    actual_delete_count = await aio_engine.remove(
        PersonModel, PersonModel.first_name == "Jean-Pierre", just_one=True
    )
    assert actual_delete_count == 1
    assert await aio_engine.count(PersonModel) == 2


@pytest.mark.usefixtures("person_persisted")
def test_sync_remove_just_one(sync_engine: SyncEngine):
    actual_delete_count = sync_engine.remove(
        PersonModel, PersonModel.first_name == "Jean-Pierre", just_one=True
    )
    assert actual_delete_count == 1
    assert sync_engine.count(PersonModel) == 2


@only_on_replica
@pytest.mark.usefixtures("person_persisted")
async def test_remove_just_one_transaction(aio_engine: AIOEngine):
    async with await aio_engine.client.start_session() as session:
        async with session.start_transaction():
            actual_delete_count = await aio_engine.remove(
                PersonModel,
                PersonModel.first_name == "Jean-Pierre",
                just_one=True,
                session=session,
            )
    assert actual_delete_count == 1
    assert await aio_engine.count(PersonModel) == 2


@only_on_replica
@pytest.mark.usefixtures("person_persisted")
def test_sync_remove_just_one_transaction(sync_engine: SyncEngine):
    with sync_engine.client.start_session() as session:
        with session.start_transaction():
            actual_delete_count = sync_engine.remove(
                PersonModel,
                PersonModel.first_name == "Jean-Pierre",
                just_one=True,
                session=session,
            )
    assert actual_delete_count == 1
    assert sync_engine.count(PersonModel) == 2


@only_on_replica
@pytest.mark.usefixtures("person_persisted")
async def test_remove_transaction_failure(aio_engine: AIOEngine):
    with pytest.raises(Exception):
        async with await aio_engine.client.start_session() as session:
            async with session.start_transaction():
                await aio_engine.remove(
                    PersonModel,
                    PersonModel.first_name == "Jean-Pierre",
                    session=session,
                )
                raise Exception("oops")
    assert await aio_engine.count(PersonModel) == 3  # type: ignore # (unreachable)


@only_on_replica
@pytest.mark.usefixtures("person_persisted")
def test_sync_remove_transaction_failure(sync_engine: SyncEngine):
    with pytest.raises(Exception):
        with sync_engine.client.start_session() as session:
            with session.start_transaction():
                sync_engine.remove(
                    PersonModel,
                    PersonModel.first_name == "Jean-Pierre",
                    session=session,
                )
                raise Exception("oops")
    assert sync_engine.count(PersonModel) == 3  # type: ignore # (unreachable)


@pytest.mark.usefixtures("person_persisted")
async def test_remove_not_existing(aio_engine: AIOEngine):
    instance = await aio_engine.find_one(
        PersonModel, PersonModel.last_name == "NotInDatabase"
    )
    assert instance is None
    deleted_count = await aio_engine.remove(
        PersonModel, PersonModel.last_name == "NotInDatabase"
    )
    assert deleted_count == 0


@pytest.mark.usefixtures("person_persisted")
def test_sync_remove_not_existing(sync_engine: SyncEngine):
    instance = sync_engine.find_one(
        PersonModel, PersonModel.last_name == "NotInDatabase"
    )
    assert instance is None
    deleted_count = sync_engine.remove(
        PersonModel, PersonModel.last_name == "NotInDatabase"
    )
    assert deleted_count == 0


async def test_modified_fields_cleared_on_document_saved(aio_engine: AIOEngine):
    instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    assert len(instance.__fields_modified__) > 0
    await aio_engine.save(instance)
    assert len(instance.__fields_modified__) == 0


def test_sync_modified_fields_cleared_on_document_saved(sync_engine: SyncEngine):
    instance = PersonModel(first_name="Jean-Pierre", last_name="Pernaud")
    assert len(instance.__fields_modified__) > 0
    sync_engine.save(instance)
    assert len(instance.__fields_modified__) == 0


async def test_modified_fields_cleared_on_nested_document_saved(aio_engine: AIOEngine):
    hachette = Publisher(name="Hachette Livre", founded=1826, location="FR")
    book = Book(title="They Didn't See Us Coming", pages=304, publisher=hachette)
    assert len(hachette.__fields_modified__) > 0
    await aio_engine.save(book)
    assert len(hachette.__fields_modified__) == 0


def test_sync_modified_fields_cleared_on_nested_document_saved(sync_engine: SyncEngine):
    hachette = Publisher(name="Hachette Livre", founded=1826, location="FR")
    book = Book(title="They Didn't See Us Coming", pages=304, publisher=hachette)
    assert len(hachette.__fields_modified__) > 0
    sync_engine.save(book)
    assert len(hachette.__fields_modified__) == 0


@pytest.fixture()
async def engine_one_person(aio_engine: AIOEngine):
    await aio_engine.save(PersonModel(first_name="Jean-Pierre", last_name="Pernaud"))


@pytest.mark.usefixtures("engine_one_person")
async def test_modified_fields_on_find(aio_engine: AIOEngine):
    instance = await aio_engine.find_one(PersonModel)
    assert instance is not None
    assert len(instance.__fields_modified__) == 0


@pytest.mark.usefixtures("engine_one_person")
def test_sync_modified_fields_on_find(sync_engine: SyncEngine):
    instance = sync_engine.find_one(PersonModel)
    assert instance is not None
    assert len(instance.__fields_modified__) == 0


@pytest.mark.usefixtures("engine_one_person")
async def test_modified_fields_on_document_change(aio_engine: AIOEngine):
    instance = await aio_engine.find_one(PersonModel)
    assert instance is not None
    instance.first_name = "Jackie"
    assert len(instance.__fields_modified__) == 1
    instance.last_name = "Chan"
    assert len(instance.__fields_modified__) == 2


@pytest.mark.usefixtures("engine_one_person")
def test_sync_modified_fields_on_document_change(sync_engine: SyncEngine):
    instance = sync_engine.find_one(PersonModel)
    assert instance is not None
    instance.first_name = "Jackie"
    assert len(instance.__fields_modified__) == 1
    instance.last_name = "Chan"
    assert len(instance.__fields_modified__) == 2


@pytest.mark.usefixtures("engine_one_person")
async def test_no_set_on_save_fetched_document(
    aio_engine: AIOEngine, sync_mock_collection
):
    instance = await aio_engine.find_one(PersonModel)
    assert instance is not None

    collection = sync_mock_collection()
    await aio_engine.save(instance)
    collection.update_one.assert_not_called()


@pytest.mark.usefixtures("engine_one_person")
def test_sync_no_set_on_save_fetched_document(
    sync_engine: SyncEngine, sync_mock_collection
):
    instance = sync_engine.find_one(PersonModel)
    assert instance is not None

    collection = sync_mock_collection()
    sync_engine.save(instance)
    collection.update_one.assert_not_called()


@pytest.mark.usefixtures("engine_one_person")
async def test_only_modified_set_on_save(aio_engine: AIOEngine, aio_mock_collection):
    instance = await aio_engine.find_one(PersonModel)
    assert instance is not None

    instance.first_name = "John"
    collection = aio_mock_collection()
    await aio_engine.save(instance)
    collection.update_one.assert_awaited_once()
    (_, set_arg), _ = collection.update_one.await_args
    assert set_arg == {"$set": {"first_name": "John"}}


@pytest.mark.usefixtures("engine_one_person")
def test_sync_only_modified_set_on_save(sync_engine: SyncEngine, sync_mock_collection):
    instance = sync_engine.find_one(PersonModel)
    assert instance is not None

    instance.first_name = "John"
    collection = sync_mock_collection()
    sync_engine.save(instance)
    collection.update_one.assert_called_once()
    (_, set_arg), _ = collection.update_one.call_args
    assert set_arg == {"$set": {"first_name": "John"}}


async def test_only_mutable_list_set_on_save(
    aio_engine: AIOEngine, aio_mock_collection
):
    class M(Model):
        field: List[str]
        immutable_field: int

    instance = M(field=["hello"], immutable_field=12)
    await aio_engine.save(instance)

    collection = aio_mock_collection()
    await aio_engine.save(instance)
    collection.update_one.assert_awaited_once()
    (_, set_arg), _ = collection.update_one.await_args
    set_dict = set_arg["$set"]
    assert list(set_dict.keys()) == ["field"]


def test_sync_only_mutable_list_set_on_save(
    sync_engine: SyncEngine, sync_mock_collection
):
    class M(Model):
        field: List[str]
        immutable_field: int

    instance = M(field=["hello"], immutable_field=12)
    sync_engine.save(instance)

    collection = sync_mock_collection()
    sync_engine.save(instance)
    collection.update_one.assert_called_once()
    (_, set_arg), _ = collection.update_one.call_args
    set_dict = set_arg["$set"]
    assert list(set_dict.keys()) == ["field"]


async def test_only_mutable_list_of_embedded_set_on_save(
    aio_engine: AIOEngine, aio_mock_collection
):
    class E(EmbeddedModel):
        a: str

    class M(Model):
        field: List[E]

    instance = M(field=[E(a="hello")])
    await aio_engine.save(instance)

    collection = aio_mock_collection()
    await aio_engine.save(instance)
    collection.update_one.assert_awaited_once()
    (_, set_arg), _ = collection.update_one.await_args
    set_dict = set_arg["$set"]
    assert set_dict == {"field": [{"a": "hello"}]}


def test_sync_only_mutable_list_of_embedded_set_on_save(
    sync_engine: SyncEngine, sync_mock_collection
):
    class E(EmbeddedModel):
        a: str

    class M(Model):
        field: List[E]

    instance = M(field=[E(a="hello")])
    sync_engine.save(instance)

    collection = sync_mock_collection()
    sync_engine.save(instance)
    collection.update_one.assert_called_once()
    (_, set_arg), _ = collection.update_one.call_args
    set_dict = set_arg["$set"]
    assert set_dict == {"field": [{"a": "hello"}]}


async def test_only_mutable_dict_of_embedded_set_on_save(
    aio_engine: AIOEngine, aio_mock_collection
):
    class E(EmbeddedModel):
        a: str

    class M(Model):
        field: Dict[str, E]

    instance = M(field={"hello": E(a="world")})
    await aio_engine.save(instance)

    collection = aio_mock_collection()
    await aio_engine.save(instance)
    collection.update_one.assert_awaited_once()
    (_, set_arg), _ = collection.update_one.await_args
    set_dict = set_arg["$set"]
    assert set_dict == {"field": {"hello": {"a": "world"}}}


def test_sync_only_mutable_dict_of_embedded_set_on_save(
    sync_engine: SyncEngine, sync_mock_collection
):
    class E(EmbeddedModel):
        a: str

    class M(Model):
        field: Dict[str, E]

    instance = M(field={"hello": E(a="world")})
    sync_engine.save(instance)

    collection = sync_mock_collection()
    sync_engine.save(instance)
    collection.update_one.assert_called_once()
    (_, set_arg), _ = collection.update_one.call_args
    set_dict = set_arg["$set"]
    assert set_dict == {"field": {"hello": {"a": "world"}}}


async def test_only_tuple_of_embedded_set_on_save(
    aio_engine: AIOEngine, aio_mock_collection
):
    class E(EmbeddedModel):
        a: str

    class M(Model):
        field: Tuple[E]

    instance = M(field=(E(a="world"),))
    await aio_engine.save(instance)

    collection = aio_mock_collection()
    await aio_engine.save(instance)
    collection.update_one.assert_awaited_once()
    (_, set_arg), _ = collection.update_one.await_args
    set_dict = set_arg["$set"]
    assert set_dict == {
        "field": [
            {"a": "world"},
        ]
    }


def test_sync_only_tuple_of_embedded_set_on_save(
    sync_engine: SyncEngine, sync_mock_collection
):
    class E(EmbeddedModel):
        a: str

    class M(Model):
        field: Tuple[E]

    instance = M(field=(E(a="world"),))
    sync_engine.save(instance)

    collection = sync_mock_collection()
    sync_engine.save(instance)
    collection.update_one.assert_called_once()
    (_, set_arg), _ = collection.update_one.call_args
    set_dict = set_arg["$set"]
    assert set_dict == {
        "field": [
            {"a": "world"},
        ]
    }


async def test_find_sort_asc(
    aio_engine: AIOEngine, person_persisted: List[PersonModel]
):
    results = await aio_engine.find(PersonModel, sort=PersonModel.last_name)
    assert results == sorted(person_persisted, key=lambda person: person.last_name)


def test_sync_find_sort_asc(
    sync_engine: SyncEngine, person_persisted: List[PersonModel]
):
    results = list(sync_engine.find(PersonModel, sort=PersonModel.last_name))
    assert results == sorted(person_persisted, key=lambda person: person.last_name)


async def test_find_sort_list(
    aio_engine: AIOEngine, person_persisted: List[PersonModel]
):
    results = await aio_engine.find(
        PersonModel, sort=(PersonModel.first_name, PersonModel.last_name)
    )
    assert results == sorted(
        person_persisted, key=lambda person: (person.first_name, person.last_name)
    )


def test_sync_find_sort_list(
    sync_engine: SyncEngine, person_persisted: List[PersonModel]
):
    results = list(
        sync_engine.find(
            PersonModel, sort=(PersonModel.first_name, PersonModel.last_name)
        )
    )
    assert results == sorted(
        person_persisted, key=lambda person: (person.first_name, person.last_name)
    )


async def test_find_sort_wrong_argument(aio_engine: AIOEngine):
    with pytest.raises(
        TypeError,
        match=(
            "sort has to be a Model field or "
            "asc, desc descriptors or a tuple of these"
        ),
    ):
        await aio_engine.find(PersonModel, sort="first_name")


def test_sync_find_sort_wrong_argument(sync_engine: SyncEngine):
    with pytest.raises(
        TypeError,
        match=(
            "sort has to be a Model field or "
            "asc, desc descriptors or a tuple of these"
        ),
    ):
        sync_engine.find(PersonModel, sort="first_name")


async def test_find_sort_wrong_tuple_argument(aio_engine: AIOEngine):
    with pytest.raises(
        TypeError,
        match="sort elements have to be Model fields or asc, desc descriptors",
    ):
        await aio_engine.find(PersonModel, sort=("first_name",))


def test_sync_find_sort_wrong_tuple_argument(sync_engine: SyncEngine):
    with pytest.raises(
        TypeError,
        match="sort elements have to be Model fields or asc, desc descriptors",
    ):
        sync_engine.find(PersonModel, sort=("first_name",))


async def test_find_sort_desc(
    aio_engine: AIOEngine, person_persisted: List[PersonModel]
):
    results = await aio_engine.find(
        PersonModel,
        sort=PersonModel.last_name.desc(),  # type: ignore
    )
    assert results == list(
        reversed(sorted(person_persisted, key=lambda person: person.last_name))
    )


def test_sync_find_sort_desc(
    sync_engine: SyncEngine, person_persisted: List[PersonModel]
):
    results = list(
        sync_engine.find(PersonModel, sort=PersonModel.last_name.desc())  # type: ignore
    )
    assert results == list(
        reversed(sorted(person_persisted, key=lambda person: person.last_name))
    )


async def test_find_sort_asc_function(
    aio_engine: AIOEngine, person_persisted: List[PersonModel]
):
    results = await aio_engine.find(PersonModel, sort=asc(PersonModel.last_name))
    assert results == sorted(person_persisted, key=lambda person: person.last_name)


def test_sync_find_sort_asc_function(
    sync_engine: SyncEngine, person_persisted: List[PersonModel]
):
    results = list(sync_engine.find(PersonModel, sort=asc(PersonModel.last_name)))
    assert results == sorted(person_persisted, key=lambda person: person.last_name)


async def test_find_sort_multiple_descriptors(aio_engine: AIOEngine):
    class TestModel(Model):
        a: int
        b: int
        c: int

    persisted_models = [
        TestModel(a=1, b=2, c=3),
        TestModel(a=2, b=2, c=3),
        TestModel(a=3, b=3, c=2),
    ]
    await aio_engine.save_all(persisted_models)
    results = await aio_engine.find(
        TestModel,
        sort=(
            desc(TestModel.a),
            TestModel.b,
            TestModel.c.asc(),  # type: ignore
        ),
    )
    assert results == sorted(
        persisted_models,
        key=lambda test_model: (-test_model.a, test_model.b, test_model.c),
    )


def test_sync_find_sort_multiple_descriptors(sync_engine: SyncEngine):
    class TestModel(Model):
        a: int
        b: int
        c: int

    persisted_models = [
        TestModel(a=1, b=2, c=3),
        TestModel(a=2, b=2, c=3),
        TestModel(a=3, b=3, c=2),
    ]
    sync_engine.save_all(persisted_models)
    results = list(
        sync_engine.find(
            TestModel,
            sort=(
                desc(TestModel.a),
                TestModel.b,
                TestModel.c.asc(),  # type: ignore
            ),
        )
    )
    assert results == sorted(
        persisted_models,
        key=lambda test_model: (-test_model.a, test_model.b, test_model.c),
    )


async def test_sort_embedded_field(aio_engine: AIOEngine):
    class E(EmbeddedModel):
        field: int

    class M(Model):
        e: E

    instances = [M(e=E(field=0)), M(e=E(field=1)), M(e=E(field=2))]
    await aio_engine.save_all(instances)
    results = await aio_engine.find(M, sort=desc(M.e.field))
    assert results == sorted(instances, key=lambda instance: -instance.e.field)


def test_sync_sort_embedded_field(sync_engine: SyncEngine):
    class E(EmbeddedModel):
        field: int

    class M(Model):
        e: E

    instances = [M(e=E(field=0)), M(e=E(field=1)), M(e=E(field=2))]
    sync_engine.save_all(instances)
    results = list(sync_engine.find(M, sort=desc(M.e.field)))
    assert results == sorted(instances, key=lambda instance: -instance.e.field)


async def test_find_one_sort(
    aio_engine: AIOEngine, person_persisted: List[PersonModel]
):
    person = await aio_engine.find_one(PersonModel, sort=PersonModel.last_name)
    assert person is not None
    assert person.last_name == "Castaldi"


def test_sync_find_one_sort(
    sync_engine: SyncEngine, person_persisted: List[PersonModel]
):
    person = sync_engine.find_one(PersonModel, sort=PersonModel.last_name)
    assert person is not None
    assert person.last_name == "Castaldi"


async def test_find_document_field_not_set_with_default(aio_engine: AIOEngine):
    class M(Model):
        field: Optional[str] = None

    await aio_engine.get_collection(M).insert_one({"_id": ObjectId()})
    gathered = await aio_engine.find_one(M)
    assert gathered is not None
    assert gathered.field is None


def test_sync_find_document_field_not_set_with_default(sync_engine: SyncEngine):
    class M(Model):
        field: Optional[str] = None

    sync_engine.get_collection(M).insert_one({"_id": ObjectId()})
    gathered = sync_engine.find_one(M)
    assert gathered is not None
    assert gathered.field is None


async def test_find_document_field_not_set_with_default_field_descriptor(
    aio_engine: AIOEngine,
):
    class M(Model):
        field: str = Field(default="hello world")

    await aio_engine.get_collection(M).insert_one({"_id": ObjectId()})
    gathered = await aio_engine.find_one(M)
    assert gathered is not None
    assert gathered.field == "hello world"


def test_sync_find_document_field_not_set_with_default_field_descriptor(
    sync_engine: SyncEngine,
):
    class M(Model):
        field: str = Field(default="hello world")

    sync_engine.get_collection(M).insert_one({"_id": ObjectId()})
    gathered = sync_engine.find_one(M)
    assert gathered is not None
    assert gathered.field == "hello world"


async def test_find_document_field_not_set_with_no_default(aio_engine: AIOEngine):
    class M(Model):
        field: str

    oid = ObjectId()
    await aio_engine.get_collection(M).insert_one({"_id": oid})
    with pytest.raises(DocumentParsingError) as exc_info:
        await aio_engine.find_one(M)
    assert redact_objectid(str(exc_info.value), oid) == snapshot(
        """\
1 validation error for M
field
  Key 'field' not found in document [type=odmantic::key_not_found_in_document, input_value={'_id': ObjectId('<ObjectId>')}, input_type=dict]\
"""  # noqa: E501
    )


def test_sync_find_document_field_not_set_with_no_default(sync_engine: SyncEngine):
    class M(Model):
        field: str

    oid = ObjectId()
    sync_engine.get_collection(M).insert_one({"_id": oid})
    with pytest.raises(DocumentParsingError) as exc_info:
        sync_engine.find_one(M)

    assert redact_objectid(str(exc_info.value), oid) == snapshot(
        """\
1 validation error for M
field
  Key 'field' not found in document [type=odmantic::key_not_found_in_document, input_value={'_id': ObjectId('<ObjectId>')}, input_type=dict]\
"""  # noqa: E501
    )


async def test_find_document_field_not_set_with_default_factory_disabled(
    aio_engine: AIOEngine,
):
    class M(Model):
        field: str = Field(default_factory=lambda: "hello")  # pragma: no cover

    await aio_engine.get_collection(M).insert_one({"_id": ObjectId()})
    with pytest.raises(DocumentParsingError, match="Key 'field' not found in document"):
        await aio_engine.find_one(M)


def test_sync_find_document_field_not_set_with_default_factory_disabled(
    sync_engine: SyncEngine,
):
    class M(Model):
        field: str = Field(default_factory=lambda: "hello")  # pragma: no cover

    sync_engine.get_collection(M).insert_one({"_id": ObjectId()})
    with pytest.raises(DocumentParsingError, match="Key 'field' not found in document"):
        sync_engine.find_one(M)


async def test_find_document_field_not_set_with_default_factory_enabled(
    aio_engine: AIOEngine,
):
    class M(Model):
        field: str = Field(default_factory=lambda: "hello")

        model_config = {"parse_doc_with_default_factories": True}

    await aio_engine.get_collection(M).insert_one({"_id": ObjectId()})
    instance = await aio_engine.find_one(M)
    assert instance is not None
    assert instance.field == "hello"


def test_sync_find_document_field_not_set_with_default_factory_enabled(
    sync_engine: SyncEngine,
):
    class M(Model):
        field: str = Field(default_factory=lambda: "hello")

        model_config = {
            "parse_doc_with_default_factories": True,
        }

    sync_engine.get_collection(M).insert_one({"_id": ObjectId()})
    instance = sync_engine.find_one(M)
    assert instance is not None
    assert instance.field == "hello"



import enum
from typing import Dict, List

from odmantic.field import Field
from odmantic.model import Model


class TreeKind(str, enum.Enum):
    BIG = "big"
    SMALL = "small"


class TreeModel(Model):
    name: str = Field(default="Acacia des montagnes")
    average_size: float = Field(key_name="size")
    discovery_year: int
    kind: TreeKind
    genesis_continents: List[str]
    per_continent_density: Dict[str, float]



import re
from datetime import datetime
from decimal import Decimal
from typing import Pattern

import pytest
import pytz
from bson.decimal128 import Decimal128
from bson.objectid import ObjectId
from bson.regex import Regex
from pydantic import ValidationError

from odmantic.bson import WithBsonSerializer
from odmantic.field import Field
from odmantic.model import Model
from odmantic.typing import Annotated

pytestmark = pytest.mark.asyncio


def test_datetime_non_naive():
    class ModelWithDate(Model):
        field: datetime

    with pytest.raises(ValueError):
        ModelWithDate(field=datetime.now(tz=pytz.timezone("Europe/Amsterdam")))

    with pytest.raises(ValueError):
        ModelWithDate(field="2018-11-02T23:59:01.824+10:00")


def test_datetime_non_naive_utc():
    class ModelWithDate(Model):
        field: datetime = Field(datetime.now(tz=pytz.utc))

    ModelWithDate()


def test_datetime_non_naive_utc_as_simplified_extended_iso_format_string():
    class ModelWithDate(Model):
        field: datetime = Field("2018-11-02T23:59:01.824Z")

    ModelWithDate()


def test_datetime_non_naive_utc_as_gmt_zero_offset_string():
    class ModelWithDate(Model):
        field: datetime = Field("2018-11-02T23:59:01.824+00:00")

    ModelWithDate()


def test_datetime_naive():
    class ModelWithDate(Model):
        field: datetime = Field(default_factory=datetime.utcnow)

    ModelWithDate()


def test_datetime_milliseconds_rounding():
    class ModelWithDate(Model):
        field: datetime

    sample_datetime = datetime.now()
    sample_datetime = sample_datetime.replace(
        microsecond=10001
    )  # Ensure we have some micro seconds that will be truncated
    inst = ModelWithDate(field=sample_datetime)

    assert inst.field.microsecond != sample_datetime.microsecond
    assert inst.field == sample_datetime.replace(microsecond=10000)

    sample_datetime = sample_datetime.replace(microsecond=999501)
    inst = ModelWithDate(field=sample_datetime)

    assert inst.field == sample_datetime.replace(microsecond=999000)


def test_validate_datetime_from_strings():
    class ModelWithDate(Model):
        field: datetime

    sample_datetime = datetime.now().replace(
        microsecond=10000
    )  # Ensure we have no micro seconds that will be truncated
    sample_datetime_str = str(sample_datetime)
    inst = ModelWithDate(field=sample_datetime_str)
    assert inst.field == sample_datetime


def test_validate_bson_objectid():
    class MyModel(Model):
        pass

    my_oid = ObjectId()
    instance = MyModel(id=str(my_oid))
    assert instance.id == my_oid


def test_validate_invalid_bson_objectid():
    class MyModel(Model):
        pass

    with pytest.raises(ValidationError) as exc_info:
        MyModel(id="not an objectid")
    errors = exc_info.value.errors()
    assert len(errors) == 2
    assert all(error["loc"][0] == "id" for error in errors)
    assert "Value error, Invalid ObjectId" in [error["msg"] for error in errors]


def test_validate_decimal_valid_string():
    class MyModel(Model):
        field: Decimal

    value = "3.152345596"
    instance = MyModel(field=value)
    assert isinstance(instance.field, Decimal)
    assert str(instance.field) == value


def test_validate_decimal_valid_bson_decimal():
    class MyModel(Model):
        field: Decimal

    str_value = "3.152345596"
    value = Decimal128(str_value)
    instance = MyModel(field=value)
    assert isinstance(instance.field, Decimal)
    assert str(instance.field) == str_value


def test_validate_decimal_invalid_string():
    class MyModel(Model):
        field: Decimal

    with pytest.raises(ValidationError) as exc_info:
        MyModel(field="3abcd.15")
    errors = exc_info.value.errors()
    assert len(errors) == 3
    assert all(error["loc"][0] == "field" for error in errors)
    assert "Value error, Invalid decimal string" in [error["msg"] for error in errors]


def test_validate_bson_decimal_valid_string():
    class MyModel(Model):
        field: Decimal128

    value = "3.152345596"
    instance = MyModel(field=value)
    assert isinstance(instance.field, Decimal128)
    assert str(instance.field) == value


def test_validate_bson_decimal_valid_bson_decimal():
    class MyModel(Model):
        field: Decimal128

    value = Decimal128("3.152345596")
    instance = MyModel(field=value)
    assert isinstance(instance.field, Decimal128)
    assert instance.field == value


def test_validate_bson_decimal_invalid_string():
    class MyModel(Model):
        field: Decimal128

    with pytest.raises(ValidationError) as exc_info:
        MyModel(field="3abcd.15")
    errors = exc_info.value.errors()
    assert len(errors) == 2
    assert all(error["loc"][0] == "field" for error in errors)
    assert "Value error, Invalid Decimal128 value" in [error["msg"] for error in errors]


def test_validate_regex_valid_regex():
    class MyModel(Model):
        field: Regex

    regex = Regex("^.*$")
    instance = MyModel(field=regex)
    assert isinstance(instance.field, Regex)
    assert instance.field == regex


def test_validate_regex_valid_string():
    class MyModel(Model):
        field: Regex

    value = "^.*$"
    instance = MyModel(field=value)
    assert isinstance(instance.field, Regex)
    assert instance.field.pattern == value


def test_validate_regex_invalid_string():
    class MyModel(Model):
        field: Regex

    with pytest.raises(ValidationError) as exc_info:
        MyModel(field="^((")
    errors = exc_info.value.errors()
    assert len(errors) == 3
    assert all(error["loc"][0] == "field" for error in errors)
    assert "Value error, Invalid Pattern value" in [error["msg"] for error in errors]


def test_validate_pattern_valid_string():
    class MyModel(Model):
        field: Pattern

    value = "^.*$"
    instance = MyModel(field=value)
    assert isinstance(instance.field, Pattern)
    assert instance.field.pattern == value


def test_validate_pattern_valid_bson_regex():
    class MyModel(Model):
        field: Pattern

    value = Regex("^.*$", flags="im")
    flags_int_value = re.compile("", flags=re.I | re.M).flags
    instance = MyModel(field=value)
    assert isinstance(instance.field, Pattern)
    assert instance.field.pattern == value.pattern
    assert instance.field.flags == flags_int_value


def test_validate_pattern_invalid_string():
    class MyModel(Model):
        field: Pattern

    with pytest.raises(ValidationError) as exc_info:
        MyModel(field="^((")
    errors = exc_info.value.errors()
    assert len(errors) == 3
    assert all(error["loc"][0] == "field" for error in errors)
    assert "Value error, Invalid Pattern value" in [error["msg"] for error in errors]


def test_with_bson_serializer_override_builtin_bson():
    MyObjectId = Annotated[ObjectId, WithBsonSerializer(lambda _: "encoded")]

    class M(Model):
        id: MyObjectId = Field(..., default_factory=ObjectId, primary_field=True)

    instance = M()
    parsed = instance.model_dump_doc()
    assert parsed == {"_id": "encoded"}



from decimal import Decimal

import bson

from odmantic import Model


def test_objectid_serialization():
    class M(Model): ...

    instance = M()
    doc = instance.model_dump_doc()
    assert isinstance(doc["_id"], bson.ObjectId)
    assert doc["_id"] == instance.id


def test_extra_allowed_bson_serialization():
    class M(Model):
        ...

        model_config = {"extra": "allow"}

    instance = M(extra_field=Decimal("1.1"))
    doc = instance.model_dump_doc()
    assert isinstance(doc["extra_field"], bson.Decimal128)



from odmantic import Model, query


class Tree(Model):
    name: str


Tree.name.match(r"^Spruce")
#> QueryExpression({'name': re.compile('^Spruce')})
query.match(Tree.name, r"^Spruce")
#> QueryExpression({'name': re.compile('^Spruce')})



from odmantic import AIOEngine, Field, Model


class Player(Model):
    name: str = Field(primary_field=True)


leeroy = Player(name="Leeroy Jenkins")
print(repr(leeroy))
#> Player(name="Leeroy Jenkins")

engine = AIOEngine()
await engine.save(leeroy)



from typing import Dict
from unittest.mock import patch

import pytest
from async_asgi_testclient import TestClient

from docs.examples_src.usage_fastapi.base_example import Tree
from odmantic.engine import AIOEngine

pytestmark = pytest.mark.asyncio


@pytest.fixture
async def base_example_client(aio_engine: AIOEngine) -> TestClient:
    with patch("docs.examples_src.usage_fastapi.base_example.engine", aio_engine):
        from docs.examples_src.usage_fastapi.base_example import app

        async with TestClient(app) as client:
            yield client


EXAMPLE_TREE_BODY = dict(name="MyTree", average_size=152, discovery_year=1992)


async def test_create_tree(base_example_client: TestClient, aio_engine: AIOEngine):
    response = await base_example_client.put("/trees/", json=EXAMPLE_TREE_BODY)
    assert response.status_code == 200
    assert await aio_engine.find_one(Tree, EXAMPLE_TREE_BODY) is not None


def is_sub_dict(a: Dict, b: Dict) -> bool:
    return set(a.items()).issubset(set(b.items()))


@pytest.mark.parametrize("count", [2, 10])
async def test_create_trees_count_get(
    base_example_client: TestClient, aio_engine: AIOEngine, count: int
):
    for _ in range(count):
        response = await base_example_client.put("/trees/", json=EXAMPLE_TREE_BODY)
        assert response.status_code == 200
    assert await aio_engine.count(Tree) == count
    async for tree in aio_engine.find(Tree):
        assert is_sub_dict(EXAMPLE_TREE_BODY, tree.model_dump())


async def test_get_tree_by_id(base_example_client: TestClient, aio_engine: AIOEngine):
    tree = Tree(**EXAMPLE_TREE_BODY)
    await aio_engine.save(tree)
    response = await base_example_client.get(
        f"/trees/{tree.id}", json=EXAMPLE_TREE_BODY
    )
    assert response.status_code == 200
    assert Tree(**response.json()) == tree


@pytest.fixture
async def example_update_client(aio_engine: AIOEngine) -> TestClient:
    with patch("docs.examples_src.usage_fastapi.example_update.engine", aio_engine):
        from docs.examples_src.usage_fastapi.example_update import app

        async with TestClient(app) as client:
            yield client


PATCHED_NAME = "New Tree Name"


async def test_update_tree_name_by_id(
    example_update_client: TestClient, aio_engine: AIOEngine
):
    tree = Tree(**EXAMPLE_TREE_BODY)
    await aio_engine.save(tree)
    response = await example_update_client.patch(
        f"/trees/{tree.id}", json=dict(name=PATCHED_NAME)
    )
    assert response.status_code == 200
    assert response.json()["name"] == PATCHED_NAME
    assert await aio_engine.find_one(Tree, {"name": PATCHED_NAME}) is not None


@pytest.fixture
async def example_delete_client(aio_engine: AIOEngine) -> TestClient:
    with patch("docs.examples_src.usage_fastapi.example_delete.engine", aio_engine):
        from docs.examples_src.usage_fastapi.example_delete import app

        async with TestClient(app) as client:
            yield client


async def test_delete_tree_by_id(
    example_delete_client: TestClient, aio_engine: AIOEngine
):
    tree = Tree(**EXAMPLE_TREE_BODY)
    await aio_engine.save(tree)
    # Create other trees not affected by the delete to come
    for _ in range(10):
        await aio_engine.save(Tree(**EXAMPLE_TREE_BODY))
    response = await example_delete_client.delete(f"/trees/{tree.id}")
    assert response.status_code == 200
    assert await aio_engine.find_one(Tree, Tree.id == tree.id) is None



from typing import List, Union

from odmantic import Model


class SimpleListModel(Model):
    field: list


print(SimpleListModel(field=[1, "a", True]).field)
#> [1, 'a', True]
print(SimpleListModel(field=(1, "a", True)).field)
#> [1, 'a', True]


class IntListModel(Model):
    field: List[int]


print(IntListModel(field=[1, 5]).field)
#> [1, 5]
print(IntListModel(field=(1, 5)).field)
#> [1, 5]


class IntStrListModel(Model):
    field: List[Union[int, str]]


print(IntStrListModel(field=[1, "b"]).field)
#> [1, 'b']
print(IntStrListModel(field=(1, "b")).field)
#> [1, 'b']



from typing import List

from bson import ObjectId

from odmantic import AIOEngine, Model


class Author(Model):
    name: str


class Book(Model):
    title: str
    pages: int
    author_ids: List[ObjectId]


david = Author(name="David Beazley")
brian = Author(name="Brian K. Jones")

python_cookbook = Book(
    title="Python Cookbook", pages=706, author_ids=[david.id, brian.id]
)
python_essentials = Book(
    title="Python Essential Reference", pages=717, author_ids=[brian.id]
)

engine = AIOEngine()
await engine.save_all((david, brian))
await engine.save_all((python_cookbook, python_essentials))



from datetime import datetime

from odmantic.bson import BSON_TYPES_ENCODERS, BaseBSONModel, ObjectId


class M(BaseBSONModel):
    id: ObjectId
    date: datetime

    model_config = {
        "json_encoders": {
            **BSON_TYPES_ENCODERS,
            datetime: lambda dt: dt.year,
        }
    }


print(M(id=ObjectId(), date=datetime.utcnow()).model_dump_json())
#> {"id": "5fa3378c8fde3766574d874d", "date": 2020}



import pytest

from odmantic.field import Field
from odmantic.model import Model
from odmantic.reference import Reference


def test_build_query_filter_across_reference():
    class Referenced(Model):
        a: int

    class M(Model):
        ref: Referenced = Reference()

    with pytest.raises(
        NotImplementedError, match="filtering across references is not supported"
    ):
        M.ref.a == 12


def test_build_query_filter_across_reference_no_attribute():
    class Referenced(Model):
        a: int

    class M(Model):
        ref: Referenced = Reference()

    with pytest.raises(AttributeError):
        M.ref.does_not_exist  # type: ignore


def test_reference_with_custom_primary_field():
    class Referenced(Model):
        key: int = Field(primary_field=True)

    class M(Model):
        ref: Referenced = Reference()

    r = Referenced(key=1)
    m = M(ref=r)
    assert m.model_dump_doc()["ref"] == 1



from odmantic import Model, SyncEngine


class Player(Model):
    name: str
    game: str


engine = SyncEngine()

with engine.session() as session:
    leeroy = session.save(Player(name="Leeroy Jenkins", game="WoW"))
    shroud = session.save(Player(name="Shroud", game="Counter-Strike"))
    with session.transaction() as transaction:
        leeroy.game = "Fortnite"
        transaction.save(leeroy)
        shroud.game = "Fortnite"
        transaction.save(shroud)
        transaction.commit()

print(engine.count(Player, Player.game == "Fortnite"))
#> 2



from odmantic.model import EmbeddedModel, Model


class Publisher(EmbeddedModel):
    name: str
    founded: int
    location: str


class Book(Model):
    title: str
    pages: int
    publisher: Publisher



import sys
from types import FunctionType
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    FrozenSet,
    List,
    Literal,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import pytest
from bson import ObjectId
from bson.decimal128 import Decimal128
from bson.regex import Regex
from pydantic import Field as PDField
from pydantic import ValidationError

from odmantic import ObjectId as ODMObjectId
from odmantic.field import Field
from odmantic.model import EmbeddedModel, Model
from odmantic.reference import Reference


class TheClassName(Model): ...


class TheClassNameModel(Model): ...


class TheClassNameOverriden(Model):
    model_config = {"collection": "collection_name"}


def test_auto_collection_name():
    assert TheClassName.__collection__ == "the_class_name"

    assert TheClassNameModel.__collection__ == "the_class_name"

    assert TheClassNameOverriden.__collection__ == "collection_name"


def test_auto_collection_name_nested():
    class theNestedClassName(Model): ...

    assert theNestedClassName.__collection__ == "the_nested_class_name"

    class TheNestedClassNameOverriden(Model):
        model_config = {"collection": "collection_name"}

    assert TheNestedClassNameOverriden.__collection__ == "collection_name"


def test_get_collection_name_pos():
    class Thing(Model): ...

    assert +Thing == "thing"


def test_duplicated_key_name():
    with pytest.raises(TypeError):

        class M(Model):
            a: int
            b: int = Field(key_name="a")


def test_duplicated_key_name_in_reference():
    class Referenced(Model):
        a: int

    with pytest.raises(TypeError):

        class Base(Model):
            a: int = Field(key_name="referenced")
            referenced: Referenced = Reference()


def test_duplicate_key_name_definition():
    with pytest.raises(TypeError):

        class Base(Model):
            a: int = Field(key_name="referenced")
            b: int = Field(key_name="referenced")


def test_key_name_containing_dollar_sign():
    class Base(Model):
        a: int = Field(key_name="a$b")


def test_key_starting_with_dollar_sign():
    with pytest.raises(TypeError):

        class Base(Model):
            a: int = Field(key_name="$a")


def test_key_containing_dot():
    with pytest.raises(TypeError):

        class Base(Model):
            b: int = Field(key_name="a.b")


def test_wrong_model_field():
    with pytest.raises(TypeError, match="use odmantic.Field instead of pydantic.Field"):

        class M(Model):
            a: int = PDField()


def test_unknown_model_field():
    class UnknownType:
        pass

    def U() -> Any:
        return UnknownType()

    with pytest.raises(TypeError):

        class M(Model):
            a: int = U()


def test_model_default_simple():
    class M(Model):
        f: int = 3

    instance = M()
    assert instance.f == 3


def test_model_default_with_field():
    class M(Model):
        f: int = Field(default=3)

    instance = M()
    assert instance.f == 3


def test_optional_field_with_default():
    class M(Model):
        f: Optional[str] = None

    assert M().f is None
    assert M(f="hello world").f == "hello world"


def test_field_with_invalid_default_type():
    with pytest.raises(TypeError, match="Unhandled field definition"):

        class M(Model):
            f: Optional[int] = "a"  # type: ignore


@pytest.mark.skip("Wait for feedback on an pydantic issue #1936")
def test_field_with_invalid_default_type_in_field():
    with pytest.raises(TypeError, match="Unhandled field definition"):

        class M(Model):
            f: Optional[int] = Field("a")


@pytest.mark.skip("Wait for feedback on an pydantic issue #1936")
def test_field_with_invalid_default_value_in_field_at_definition():
    with pytest.raises(TypeError, match="Unhandled field definition"):

        class M(Model):
            f: Optional[int] = Field(3, gt=5)


def test_field_with_invalid_default_value_in_field_at_instantiation():
    class M(Model):
        f: Optional[int] = Field(3, gt=5)

    with pytest.raises(ValidationError):
        M()


def test_optional_field_with_field_settings():
    class M(Model):
        f: Optional[str] = Field("hello world", key_name="my_field")

    assert M().f == "hello world"
    assert M(f=None).f is None


def test_unable_to_generate_primary_field():
    with pytest.raises(TypeError, match="can't automatically generate a primary field"):

        class A(Model):
            id: str


def test_define_alternate_primary_key():
    class M(Model):
        name: str = Field(primary_field=True)

    instance = M(name="Jack")
    assert instance.model_dump_doc() == {"_id": "Jack"}


def test_weird_overload_id_field():
    class M(Model):
        id: int
        name: str = Field(primary_field=True)

    instance = M(id=15, name="Johnny")
    assert instance.model_dump_doc() == {"_id": "Johnny", "id": 15}


@pytest.mark.skip("Not implemented, see if it should be supported...")
def test_overload_id_with_another_primary_key():
    with pytest.raises(TypeError, match="cannot define multiple primary keys"):

        class M(Model):
            id: int
            number: int = Field(primary_key=True)


def test_untyped_field_definition():
    with pytest.raises(TypeError, match="defined without type annotation"):

        class M(Model):
            a = 3


def test_multiple_primary_key():
    with pytest.raises(TypeError, match="Duplicated key_name"):

        class M(Model):
            a: int = Field(primary_field=True)
            b: int = Field(primary_field=True)


def test_model_with_implicit_reference_error():
    class A(Model):
        pass

    with pytest.raises(TypeError, match="without a Reference assigned"):

        class B(Model):
            a: A


def test_embedded_model_with_primary_key():
    with pytest.raises(TypeError, match="cannot define a primary field"):

        class A(EmbeddedModel):
            f: int = Field(primary_field=True)


T = TypeVar("T")


@pytest.mark.parametrize("generic", [List, Set, Tuple])
def test_embedded_model_generics_as_primary_key(generic: Type):
    class E(EmbeddedModel):
        f: int

    with pytest.raises(
        TypeError,
        match="Declaring a generic type of embedded models as a primary field"
        " is not allowed",
    ):

        class M(Model):
            e: generic[E] = Field(primary_field=True)  # type: ignore


@pytest.mark.parametrize(
    "generic",
    [
        lambda e: List[e],  # type: ignore
        lambda e: Set[e],  # type: ignore
        lambda e: Dict[str, e],  # type: ignore
        lambda e: Tuple[e],
        lambda e: Tuple[e, ...],
    ],
)
def test_embedded_model_generics_with_references(generic: Callable[[Type], Type]):
    class AnotherModel(Model):
        a: float

    class E(EmbeddedModel):
        f: AnotherModel = Reference()

    with pytest.raises(
        TypeError,
        match="Declaring a generic type of embedded models containing references"
        " is not allowed",
    ):

        class M(Model):
            e: generic(E)  # type: ignore


def test_invalid_collection_name_dollar():
    with pytest.raises(TypeError, match=r"cannot contain '\$'"):

        class A(Model):
            model_config = {"collection": "hello$world"}


def test_invalid_collection_name_empty():
    with pytest.raises(TypeError, match="cannot be empty"):

        class A(Model):
            model_config = {"collection": ""}


def test_invalid_collection_name_contain_system_dot():
    with pytest.raises(TypeError, match="cannot start with 'system.'"):

        class A(Model):
            model_config = {"collection": "system.hi"}


def test_custom_collection_name():
    class M(Model):
        model_config = {"collection": "collection_name"}

    assert M.__collection__ == "collection_name"


def test_embedded_model_key_name():
    class E(EmbeddedModel):
        f: int = 3

    class M(Model):
        field: E = Field(E(), key_name="hello")

    doc = M().model_dump_doc()
    assert "hello" in doc
    assert doc["hello"] == {"f": 3}


def test_embedded_model_as_primary_field():
    class E(EmbeddedModel):
        f: int

    class M(Model):
        field: E = Field(primary_field=True)

    assert M(field=E(f=1)).model_dump_doc() == {"_id": {"f": 1}}


def test_untouched_types_function():
    def id_str(self) -> str:  # pragma: no cover
        return str(self.id)

    class M(Model):
        model_config = {"arbitrary_types_allowed": True}

        id_: FunctionType = id_str  # type: ignore

    assert "id_" not in M.__odm_fields__.keys()


@pytest.mark.parametrize(
    "t",
    [
        Optional[ObjectId],
        List[ObjectId],
        List[Decimal128],
        List[Regex],
        FrozenSet[Regex],
        Union[Regex, ObjectId],
        Dict[ObjectId, str],
        Dict[Tuple[ObjectId, ...], str],
        Dict[Union[ObjectId, str], str],
        Mapping[Union[ObjectId, str], str],
    ],
)
def test_compound_bson_field(t: Type):
    class M(Model):
        children: t  # type: ignore


def test_forbidden_field():
    with pytest.raises(TypeError, match="fields are not supported"):

        class M(Model):
            children: Callable


def test_model_with_class_var():
    class M(Model):
        cls_var: ClassVar[str] = "theclassvar"
        field: int

    assert M.cls_var == "theclassvar"
    m = M(field=5)
    assert m.cls_var == "theclassvar"
    assert m.field == 5
    assert "cls_var" not in m.model_dump_doc().keys()


def test_model_definition_extra_allow():
    class M(Model):
        model_config = {"extra": "allow"}

        f: int

    instance = M(f=1, g=2)
    assert instance.model_dump_doc(include={"f", "g"}) == {"f": 1, "g": 2}


def test_model_definition_extra_ignore():
    class M(Model):
        model_config = {"extra": "ignore"}

        f: int

    instance = M(f=1, g=2)
    assert instance.model_dump_doc(include={"f", "g"}) == {"f": 1}


def test_model_definition_extra_forbid():
    class M(Model):
        model_config = {"extra": "forbid"}

        f: int

    with pytest.raises(ValidationError, match="Extra inputs are not permitted"):
        M(f=1, g=2)


def test_extra_field_type_subst():
    class M(Model):
        model_config = {"extra": "allow"}

        f: int

    instance = M(f=1, oid=ODMObjectId())

    assert isinstance(instance.model_dump_doc()["oid"], ObjectId)


def test_extra_field_document_parsing():
    class M(Model):
        model_config = {"extra": "allow"}

        f: int

    instance = M.model_validate_doc({"_id": ObjectId(), "f": 1, "extra": "hello"})

    assert "extra" in instance.model_dump_doc()


class EmForGenericDefinitionTest(EmbeddedModel):
    f: int


@pytest.mark.skipif(
    sys.version_info[:3] < (3, 9, 0),
    reason="Standard collection generics not supported by python < 3.9",
)
@pytest.mark.parametrize(
    "get_type, value",
    [
        (lambda: list[int], [1, 2, 3]),
        (lambda: dict[str, int], {"a": 1, "b": 2}),
        (lambda: set[int], {1, 2, 3}),
        (lambda: tuple[int, ...], (1, 2, 3)),
        (
            lambda: list[EmForGenericDefinitionTest],
            [EmForGenericDefinitionTest(f=1), EmForGenericDefinitionTest(f=2)],
        ),
        (
            lambda: dict[str, EmForGenericDefinitionTest],
            {
                "a": EmForGenericDefinitionTest(f=1),
                "b": EmForGenericDefinitionTest(f=2),
            },
        ),
        (
            lambda: tuple[EmForGenericDefinitionTest, ...],
            (EmForGenericDefinitionTest(f=1), EmForGenericDefinitionTest(f=2)),
        ),
    ],
)
def test_model_definition_with_new_generics(get_type: Callable, value: Any):
    class M(Model):
        f: get_type()  # type: ignore # 3.9 + syntax

    assert M(f=value).f == value


def test_model_definition_with_literal():
    class M(Model):
        f: Literal["a", "b", "c"]  # noqa: F821

    assert M(f="a").f == "a"


def test_model_definition_with_literal_fail():
    class M(Model):
        f: Literal["a", "b", "c"]  # noqa: F821

    with pytest.raises(ValidationError):
        M(f="w")


def test_model_definition_with_generic_literals():
    class M(Model):
        f: List[Literal["a", "b", "c"]]  # noqa: F821

    assert M(f=["a", "c"]).f == ["a", "c"]


def test_model_with_multiple_optional_fields():
    class Person(Model):
        hashed_password: Optional[str]
        totp_secret: Optional[str] = Field(default=None)
        totp_counter: Optional[int] = Field(default=None)

    user = {
        "hashed_password": "hashed_password",
    }
    Person(**user)



from typing import Tuple

from odmantic import Model


class SimpleTupleModel(Model):
    field: tuple


print(SimpleTupleModel(field=[1, "a", True]).field)
#> (1, 'a', True)
print(SimpleTupleModel(field=(1, "a", True)).field)
#> (1, 'a', True)


class TwoIntTupleModel(Model):
    field: Tuple[int, int]


print(SimpleTupleModel(field=(1, 10)).field)
#> (1, 10)
print(SimpleTupleModel(field=[1, 10]).field)
#> (1, 10)


class IntTupleModel(Model):
    field: Tuple[int, ...]


print(IntTupleModel(field=(1,)).field)
#> (1,)
print(IntTupleModel(field=[1, 2, 3, 10]).field)
#> (1, 2, 3, 10)



import sys
from typing import (
    Callable,
    Dict,
    FrozenSet,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

import bson
import pytest
from typing_utils import are_generics_equal

from odmantic.bson import (
    _BSON_SUBSTITUTED_FIELDS,
    Binary,
    Decimal128,
    Int64,
    ObjectId,
    Regex,
)
from odmantic.model import EmbeddedModel, Model, is_type_mutable, validate_type


@pytest.mark.parametrize("base, replacement", _BSON_SUBSTITUTED_FIELDS.items())
def test_validate_type_bson_substituted(base, replacement):
    assert validate_type(base) == replacement


@pytest.mark.parametrize("base, replacement", _BSON_SUBSTITUTED_FIELDS.items())
def test_optional_bson_subst(base, replacement):
    assert are_generics_equal(
        validate_type(Optional[base]),
        Optional[replacement],
    )


@pytest.mark.parametrize("origin", (List, Set, FrozenSet, Sequence))
@pytest.mark.parametrize("base, replacement", _BSON_SUBSTITUTED_FIELDS.items())
def test_single_arg_type_bson_subst(origin, base, replacement):
    assert are_generics_equal(validate_type(origin[base]), origin[replacement])


def test_forbidden_field():
    with pytest.raises(TypeError, match="fields are not supported"):
        validate_type(Callable)  # type: ignore


def test_deep_nest_bson_subst():
    assert are_generics_equal(
        validate_type(
            Union[  # type: ignore
                Dict[List[bson.ObjectId], Dict[bson.ObjectId, bson.Decimal128]],
                Dict[Dict[Set[bson.Int64], bson.Binary], Tuple[bson.Regex, ...]],
            ]
        ),
        Union[
            Dict[List[ObjectId], Dict[ObjectId, Decimal128]],
            Dict[Dict[Set[Int64], Binary], Tuple[Regex, ...]],
        ],
    )


class DummyEmbedded(EmbeddedModel):
    field: str


class DummyModel(Model):
    field: str


@pytest.mark.parametrize(
    "t",
    (
        None,
        bool,
        int,
        str,
        Tuple,
        Tuple[int, str, bool],
        Tuple[int, ...],
        FrozenSet[int],
        Union[FrozenSet[int], Tuple[int, str]],
        DummyModel,
    ),
)
def test_mutable_types_immutables(t: Type):
    assert not is_type_mutable(t)


TEST_TYPES = [
    List,
    Set,
    List[int],
    Tuple[List[int]],
    FrozenSet[Set[int]],
    Dict[Tuple[int, ...], str],
    DummyEmbedded,
    Tuple[DummyEmbedded, ...],
    Dict[str, DummyEmbedded],
    FrozenSet[DummyEmbedded],
]

# if generic with builtin types are supported add them to the list
if sys.version_info >= (3, 9):
    TEST_TYPES += [
        list,
        set,
        list[int],
        tuple[list[int]],
        frozenset[set[int]],
        dict[tuple[int, ...], str],
    ]


@pytest.mark.parametrize(
    "t",
    TEST_TYPES,
)
def test_mutable_types_mutables(t: Type):
    assert is_type_mutable(t)


def test_mutable_types_unknown_type():
    class T: ...

    assert is_type_mutable(T)


def test_mutable_field_embedded_model():
    class E(EmbeddedModel):
        f: int

    class M(Model):
        e: E

    assert "e" in M.__mutable_fields__



from odmantic import Model, SyncEngine


class Player(Model):
    name: str
    game: str


engine = SyncEngine()

for player in engine.find(Player, Player.game == "Starcraft"):
    print(repr(player))

#> Player(id=ObjectId(...), name='TLO', game='Starcraft')
#> Player(id=ObjectId(...), name='Serral', game='Starcraft')



from __future__ import annotations

import abc
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Optional,
    Pattern,
    Set,
    Type,
    Union,
    cast,
)

from pydantic.config import JsonDict
from pydantic.fields import Field as PDField
from pydantic.fields import FieldInfo, PydanticUndefined

from odmantic.config import ODMConfigDict
from odmantic.query import (
    QueryExpression,
    SortExpression,
    asc,
    desc,
    eq,
    gt,
    gte,
    in_,
    lt,
    lte,
    match,
    ne,
    not_in,
)

if TYPE_CHECKING:
    from odmantic.model import EmbeddedModel, Model  # noqa: F401

    from .typing import NoArgAnyCallable


def Field(
    default: Any = PydanticUndefined,
    *,
    key_name: Optional[str] = None,
    primary_field: bool = False,
    index: bool = False,
    unique: bool = False,
    default_factory: Optional["NoArgAnyCallable"] = None,
    # alias: str = None, # FIXME not supported yet
    title: Optional[str] = None,
    description: Optional[str] = None,
    json_schema_extra: JsonDict | Callable[[JsonDict], None] | None = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
    examples: list[Any] | None = None,
) -> Any:
    """Used to provide extra information about a field, either for the model schema or
    complex validation. Some arguments apply only to number fields (``int``, ``float``,
     ``Decimal``) and some apply only to ``str``.

    Tip:
        The main additions of ODMantic to the regular pydantic `Field` are the
        `key_name`, `index`, `unique` and the `primary_field` options.

    Warning:
        If both `default` and `default_factory` are set, an error is raised.

    Warning:
        `primary_field` can't be used along with `key_name` since the key_name will be
        set to `_id`.


    Args:
        default: since this is replacing the field’s default, its first argument is
            used to set the default, use ellipsis (``...``) to indicate the field has no
            default value
        key_name: the name to use in the the mongo document structure
        primary_field: this field should be considered as a primary key.
        index: this field should be considered as an index
        unique: this field should be considered as a unique index
        default_factory: callable that will be called when a default value is needed
            for this field.
        title: can be any string, used in the schema
        description: can be any string, used in the schema
        examples: can be any list, used in the schema
        json_schema_extra: Any additional JSON schema data for the schema property.
        const: this field is required and *must* take it's default value
        gt: only applies to numbers, requires the field to be "greater than". The
            schema will have an ``exclusiveMinimum`` validation keyword
        ge: only applies to numbers, requires the field to be "greater than or equal
            to". The schema will have a ``minimum`` validation keyword
        lt: only applies to numbers, requires the field to be "less than". The schema
            will have an ``exclusiveMaximum`` validation keyword
        le: only applies to numbers, requires the field to be "less than or equal to"
            . The schema will have a ``maximum`` validation keyword
        multiple_of: only applies to numbers, requires the field to be "a multiple of
            ". The schema will have a ``multipleOf`` validation keyword
        min_items: only applies to sequences, requires the field to have a minimum
            item count.
        max_items: only applies to sequences, requires the field to have a maximum
            item count.
        min_length: only applies to strings, requires the field to have a minimum
            length. The schema will have a ``maximum`` validation keyword
        max_length: only applies to strings, requires the field to have a maximum
            length. The schema will have a ``maxLength`` validation keyword
        regex: only applies to strings, requires the field match agains a regular
            expression pattern string. The schema will have a ``pattern`` validation
            keyword

    <!---
    # noqa: DAR201
    # noqa: DAR003
    # noqa: DAR401
    # noqa: DAR101
    -->
    """
    # Perform casts on optional fields to avoid incompatibility due to the strict
    # optional mypy setting
    # TODO: add remaining validation fields from pydantic
    pydantic_field = PDField(
        default,
        default_factory=default_factory,
        # alias=alias,  # FIXME check aliases compatibility
        title=cast(str, title),
        description=cast(str, description),
        examples=examples,
        json_schema_extra=json_schema_extra,
        const=cast(bool, const),
        gt=cast(float, gt),
        ge=cast(float, ge),
        lt=cast(float, lt),
        le=cast(float, le),
        multiple_of=cast(float, multiple_of),
        min_items=cast(int, min_items),
        max_items=cast(int, max_items),
        min_length=cast(int, min_length),
        max_length=cast(int, max_length),
        regex=cast(str, regex),
    )
    if primary_field:
        if key_name is not None and key_name != "_id":
            raise ValueError(
                "cannot specify a primary field with a custom key_name,"
                "key_name='_id' enforced"
            )
        else:
            key_name = "_id"
    elif key_name == "_id":
        raise ValueError(
            "cannot specify key_name='_id' without defining the field as primary"
        )

    return ODMFieldInfo(
        pydantic_field_info=pydantic_field,
        primary_field=primary_field,
        key_name=key_name,
        index=index,
        unique=unique,
    )


class ODMFieldInfo:
    """Extra data for an ODM field."""

    __slots__ = ("pydantic_field_info", "primary_field", "key_name", "index", "unique")

    def __init__(
        self,
        *,
        pydantic_field_info: FieldInfo,
        primary_field: bool,
        key_name: Optional[str],
        index: bool,
        unique: bool,
    ):
        self.pydantic_field_info = pydantic_field_info
        self.primary_field = primary_field
        self.key_name = key_name
        self.index = index
        self.unique = unique


class ODMBaseField(metaclass=abc.ABCMeta):
    __slots__ = ("key_name", "model_config", "pydantic_field")
    __allowed_operators__: Set[str]

    def __init__(self, key_name: str, model_config: ODMConfigDict):
        self.key_name = key_name
        self.model_config = model_config

    def bind_pydantic_field(self, field: FieldInfo) -> None:
        self.pydantic_field = field

    def is_required_in_doc(self) -> bool:
        if self.model_config["parse_doc_with_default_factories"]:
            return self.pydantic_field.is_required()
        else:
            return (
                self.pydantic_field.default_factory is not None
                or self.pydantic_field.is_required()
            )


class ODMBaseIndexableField(ODMBaseField, metaclass=abc.ABCMeta):
    __slots__ = ("index", "unique")

    def __init__(
        self,
        key_name: str,
        model_config: ODMConfigDict,
        index: bool,
        unique: bool,
    ):
        super().__init__(key_name, model_config)
        self.index = index
        self.unique = unique


class ODMField(ODMBaseIndexableField):
    """Used to interact with the ODM model class."""

    __slots__ = ("primary_field",)
    __allowed_operators__ = set(
        ("eq", "ne", "in_", "not_in", "lt", "lte", "gt", "gte", "match", "asc", "desc")
    )

    def __init__(
        self,
        *,
        key_name: str,
        model_config: ODMConfigDict,
        primary_field: bool,
        index: bool = False,
        unique: bool = False,
    ):
        super().__init__(key_name, model_config, index, unique)
        self.primary_field = primary_field

    def get_default_importing_value(self) -> Any:
        # The default importing value doesn't consider the default_factory setting by
        # default as it could result in inconsistent behaviors for datetime.now
        # factories for example
        return self.pydantic_field.get_default(
            call_default_factory=self.model_config["parse_doc_with_default_factories"]
        )


class ODMReference(ODMBaseField):
    """Field pointing on a referenced model."""

    __slots__ = ("model",)
    __allowed_operators__ = set(("eq", "ne", "in_", "not_in"))

    def __init__(
        self, key_name: str, model_config: ODMConfigDict, model: Type["Model"]
    ):
        super().__init__(key_name, model_config)
        self.model = model


class ODMEmbedded(ODMField):
    __slots__ = "model"
    __allowed_operators__ = set(("eq", "ne", "in_", "not_in"))

    def __init__(
        self,
        primary_field: bool,
        key_name: str,
        model_config: ODMConfigDict,
        model: Type["EmbeddedModel"],
        index: bool = False,
        unique: bool = False,
    ):
        super().__init__(
            primary_field=primary_field,
            key_name=key_name,
            model_config=model_config,
            index=index,
            unique=unique,
        )
        self.model = model


class ODMEmbeddedGeneric(ODMField):
    # Only dict,set and list are "officially" supported for now
    __slots__ = ("model", "generic_origin")
    __allowed_operators__ = set(("eq", "ne"))

    def __init__(
        self,
        key_name: str,
        model_config: ODMConfigDict,
        model: Type["EmbeddedModel"],
        generic_origin: Any,
        index: bool = False,
        unique: bool = False,
    ):
        super().__init__(
            primary_field=False,
            key_name=key_name,
            model_config=model_config,
            index=index,
            unique=unique,
        )
        self.model = model
        self.generic_origin = generic_origin


class KeyNameProxy(str):
    """Used to provide the `++` operator enabling reference key name creation"""

    def __pos__(self) -> str:
        return f"${self}"


class FieldProxy:
    __slots__ = ("parent", "field")

    def __init__(self, parent: Optional["FieldProxy"], field: ODMBaseField) -> None:
        self.parent = parent
        self.field = field

    def _get_key_name(self) -> str:
        parent: Optional[FieldProxy] = object.__getattribute__(self, "parent")
        field: ODMBaseField = object.__getattribute__(self, "field")

        if parent is None:
            return field.key_name

        parent_name: str = object.__getattribute__(parent, "_get_key_name")()
        return f"{parent_name}.{field.key_name}"

    def __getattribute__(self, name: str) -> Any:
        if name == "__class__":  # support `isinstance` for python < 3.7
            return super().__getattribute__(name)

        field: ODMBaseField = object.__getattribute__(self, "field")
        if isinstance(field, ODMReference):
            if name in field.model.__odm_fields__:
                raise NotImplementedError(
                    "filtering across references is not supported"
                )
        elif isinstance(field, ODMEmbedded):
            child_field = field.model.__odm_fields__.get(name)
            if child_field is None:
                try:
                    return super().__getattribute__(name)
                except AttributeError:
                    raise AttributeError(
                        f"attribute {name} not found in {field.model.__name__}"
                    )
            return FieldProxy(parent=self, field=child_field)

        if name not in field.__allowed_operators__:
            raise AttributeError(
                f"operator {name} not allowed for {type(field).__name__} fields"
            )
        return super().__getattribute__(name)

    def __pos__(self) -> KeyNameProxy:
        return KeyNameProxy(object.__getattribute__(self, "_get_key_name")())

    def __gt__(self, value: Any) -> QueryExpression:
        return self.gt(value)

    def gt(self, value: Any) -> QueryExpression:
        return gt(self, value)

    def gte(self, value: Any) -> QueryExpression:
        return gte(self, value)

    def __ge__(self, value: Any) -> QueryExpression:
        return self.gte(value)

    def lt(self, value: Any) -> QueryExpression:
        return lt(self, value)

    def __lt__(self, value: Any) -> QueryExpression:
        return self.lt(value)

    def lte(self, value: Any) -> QueryExpression:
        return lte(self, value)

    def __le__(self, value: Any) -> QueryExpression:
        return self.lte(value)

    def eq(self, value: Any) -> QueryExpression:
        return eq(self, value)

    def __eq__(self, value: Any) -> QueryExpression:  # type: ignore
        return self.eq(value)

    def ne(self, value: Any) -> QueryExpression:
        return ne(self, value)

    def __ne__(self, value: Any) -> QueryExpression:  # type: ignore
        return self.ne(value)

    def in_(self, value: Iterable) -> QueryExpression:
        return in_(self, value)

    def not_in(self, value: Iterable) -> QueryExpression:
        return not_in(self, value)

    def match(self, pattern: Union[Pattern, str]) -> QueryExpression:
        return match(self, pattern)

    def asc(self) -> SortExpression:
        return asc(self)

    def desc(self) -> SortExpression:
        return desc(self)



import pytest

from odmantic import AIOEngine
from odmantic.engine import SyncEngine
from tests.zoo.player import Player
from tests.zoo.twitter_user import TwitterUser

pytestmark = pytest.mark.asyncio


async def test_twitter_user(aio_engine: AIOEngine):
    main = TwitterUser()
    await aio_engine.save(main)
    friends = [TwitterUser() for _ in range(25)]
    await aio_engine.save_all(friends)
    friend_ids = [f.id for f in friends]
    main.following = friend_ids
    await aio_engine.save(main)

    fetched_main = await aio_engine.find_one(TwitterUser, TwitterUser.id == main.id)
    assert fetched_main is not None
    assert fetched_main == main
    assert set(friend_ids) == set(fetched_main.following)


def test_sync_twitter_user(sync_engine: SyncEngine):
    main = TwitterUser()
    sync_engine.save(main)
    friends = [TwitterUser() for _ in range(25)]
    sync_engine.save_all(friends)
    friend_ids = [f.id for f in friends]
    main.following = friend_ids
    sync_engine.save(main)

    fetched_main = sync_engine.find_one(TwitterUser, TwitterUser.id == main.id)
    assert fetched_main is not None
    assert fetched_main == main
    assert set(friend_ids) == set(fetched_main.following)


async def test_player(aio_engine: AIOEngine):
    leeroy = Player(name="Leeroy Jenkins")
    await aio_engine.save(leeroy)
    fetched = await aio_engine.find_one(Player)
    assert fetched == leeroy


def test_sync_player(sync_engine: SyncEngine):
    leeroy = Player(name="Leeroy Jenkins")
    sync_engine.save(leeroy)
    fetched = sync_engine.find_one(Player)
    assert fetched == leeroy



from odmantic import SyncEngine, Model


class Player(Model):
    name: str
    game: str


engine = SyncEngine()
shroud = engine.find_one(Player, Player.name == "Shroud")
print(shroud.game)
#> Counter-Strike
shroud.game = "Valorant"
engine.save(shroud)



from typing import Dict, List, Set, Tuple

from typing_utils import are_generics_equal


def test_are_generics_equal_two_different_origin():
    assert not are_generics_equal(List[str], Set[str])


def test_are_generics_equal_different_arg_count():
    assert not are_generics_equal(Tuple[str], Tuple[str, str])
    assert not are_generics_equal(Tuple[str], Tuple[str, ...])


def test_are_generics_equal_different_args():
    assert not are_generics_equal(Tuple[str, int], Tuple[int, str])
    assert not are_generics_equal(
        Tuple[str, int, Dict[int, str]], Tuple[str, int, Dict[int, int]]
    )



from odmantic.model import Model
from odmantic.reference import Reference


class Publisher(Model):
    name: str
    founded: int
    location: str


class Book(Model):
    title: str
    pages: int
    publisher: Publisher = Reference()



from odmantic.model import Model
from odmantic.reference import Reference


class NestedLevel3(Model):
    field: int = 3


class NestedLevel2(Model):
    field: int = 2
    next_: NestedLevel3 = Reference()


class NestedLevel1(Model):
    field: int = 1
    next_: NestedLevel2 = Reference()
